"""
This module contains the base code and interface of basic Taskbot.
"""
from botiverse.bots.BasicTaskBot.utils import RuleBasedNLU, MostRecentDST, RandomDP, TemplateBasedNLG


class BasicTaskBot:
    """
    Instantiate a Task Oriented Dialogue System chat bot.
    It aims to assist the user in completing certain tasks in specific domains.
    The chat bot can use a fully Rule-Based approach which required no data for
    training and can also include Deep learning and data can be then used to train
    the chatbot model.

    :param domains_slots: The slots of each domain in the system.
    :type domains_slots: dict[str, list[str]]

    :param templates: The predefined templates, where templates["domain"]["slot"] is
        a list of questions asking about slot "slot" in domain "domain".
    :type templates: dict[str, dict[str, list[str]]]

    :param domains_pattern: Dictionary where the key is a name of a domain
        and the value is a Regex that is used to capture that domain.
    :type domains_pattern: dict[str, str]

    :param slots_pattern: Dictionary where the key is a name of a domain
        and the value is another dictionary where the key is a name of a slot
        inside the domain and the value is a Regex that is used to capture that slot.
    :type slots_pattern: dict[str, dict[str, str]]

    """

    def __init__(self, domains_slots, templates, domains_pattern, slots_pattern, verbose=False):
        self.nlu = RuleBasedNLU(domains_pattern, slots_pattern)
        self.dst = MostRecentDST(domains_slots)
        self.dpo = RandomDP()
        self.nlg = TemplateBasedNLG(templates)
        self.verbose = verbose
        self.current_domain = None

    def read_data(self, data):
        """
        Read the chatbot training data, redundant here as we are using a fully classical approach.

        :param data: any
        :type number: any

        :return: None
        :rtype: NoneType
        """
        pass

    def train(self):
        """
        Train the chatbot model with the given data in read_data, redundant here as we are using a fully classical approach.

        :param data: any
        :type number: any

        :return: None
        :rtype: NoneType
        """
        pass
    
    def infer(self, prompt):
        """
        Infer a suitable response to the given prompt.

        :param promp: The user's prompt
        :type number: str

        :return: When all slots in the current domain are filled it returns empty string, otherwise it returns the response.
        :rtype: str
        """

        response = ""
        status, self.current_domain = self.nlu.get_domain(self.current_domain, prompt)
        # if no domain is detected or the domain is not supported return a default response
        if status is False or self.current_domain is None:
            response = "Sorry I don't know what domain are you talking about!"
        else:
            # get the slots and their values
            status, slots, values = self.nlu.get_slot_fillers(self.current_domain, prompt)
            # if no slots are detected or the slots are not supported return a default response
            if status is False:
                response = "Sorry I Couldn't catch that can you say it in a different way!"
            else:
                # update the dialogue state
                self.dst.update_state(self.current_domain, slots, values)
                # get the action to be taken
                action = self.dpo.get_action(self.current_domain, self.dst.get_dialogue_state())
                # if no action is detected return empty response
                if action == "":
                    response = ""
                else:
                    response = self.nlg.generate(self.current_domain, action)

        if self.verbose == True:
            print(f'State: {self.dst.get_dialogue_state()}')

        return response

    def get_dialogue_state(self):
        """
        Gets the dialogue state.

        :return: The dialogue state, where state["domain"]["slot"] indicates
            value of slot "slot" in domain "domain".
        :rtype: dict[str, dict[str, str]]
        """
        return self.dst.get_dialogue_state()

    def reset(self, domain=None, slot=None):
        """
        Reset the dialogue state.
        Note: if the domain or the slot is equal to None all domains
        or slots will be reset respectively.

        :param domain: The domain to be reset, defaults to None.
        :type domain: str

        :param slot: The slot to be reset, defaults to None.
        :type slot: str

        """
        self.dst.reset(domain, slot)



try:
    import numpy as np
    import json
    from gtts import gTTS
    import tempfile
    import os
    from botiverse.models import TTS
    from playsound import playsound
    from botiverse.models import LSTMClassifier
    from botiverse.preprocessors import Vocalize, Wav2Vec, Wav2Text, BertEmbedder, Frequency, BertSentenceEmbedder
    from botiverse.bots.VoiceBot.utils import voice_input
except:
    pass

class VoiceBot():
    '''An interface for the vocalizer chatbot which simulates a call with a customer service bot.'''
    def __init__(self,  call_json_path, repr='BERT-Sentence'):
        ''' 
        Load the call data from a json file.
        :param call_json_path: The path to the json file containing the call state machine.
        '''
        with open(call_json_path, 'r') as file:
            call_json = file.read()
        self.call_data = json.loads(call_json)
        self.current_node = 'A'
        self.wav2text = Wav2Text()
        if repr == 'BERT':
            self.bert_embeddings = BertEmbedder()
        elif repr == 'BERT-Sentence':
            self.bert_sentence_embeddings = BertSentenceEmbedder()
        else:
            raise Exception(f"Invalid representation {repr}. Expected BERT or BERT-Sentence.")        
    
    def generate_speech(self, text, offline=False):
        '''Use Google's TTS or offline FastSpeech 1.0 to play speech from the given text.
        :param text: The text to be converted into speech.
        :param offline: Whether to use offline FastSpeech 1.0 to generate speech.
        '''
        if offline:
            tts = TTS()
            tts.speak(text)
        else:
            tts = gTTS(text=text, lang='en', tld="us", slow=False)
            with tempfile.NamedTemporaryFile(suffix=".mp3", delete=False) as temp_audio:
                temp_filename = temp_audio.name
                tts.save(temp_filename)
                playsound(temp_filename)

    def simulate_call(self):
        '''
        Simulate a call with a customer service bot as driven by the call state machine.
        '''
        while True:
            if self.current_node == 'Z':
                # the final state has a different structure, bot only speaks and then the call ends
                bot_message = self.call_data[self.current_node]['Bot']
                self.generate_speech(bot_message)
                break

            # 1 - get the current node's data and from that get the message the bot should speak
            node_data = self.call_data[self.current_node]
            bot_message = node_data['Bot']
            self.generate_speech(bot_message)

            # 2 - get the intent options that the bot expects from the user and classify the user's response
            options = node_data['Options']
            intents = [option['Intent'] for option in options]
            max_dur = node_data['max_duration']
            human_resp = voice_input(record_time=int(max_dur))
            human_resp = self.wav2text.transcribe(human_resp)
            selected_ind, score = self.bert_sentence_embeddings.closest_sentence(human_resp, intents, retun_ind=True)
            print(f"you said: {human_resp} and the bot decided that you meant {intents[selected_ind]} with a score of {score}")
            
            # 3 - speak according to the chosen option
            speak_message = options[selected_ind]['Speak']
            self.generate_speech(speak_message)

            # 4 - go to the next state
            self.current_node = options[selected_ind]['Next']




"""
This Module has base code and interfaces for TripPy Dialogue State Tracker.
"""

import torch
from collections import OrderedDict
import gdown
import os

from botiverse.models.TRIPPY.utils import normalize, mask_utterance
from botiverse.models.TRIPPY.data import fix_slot_list, prepare_data, Dataset
from botiverse.models.TRIPPY.run import run
from botiverse.models.TRIPPY.infer import infer
from botiverse.models.TRIPPY.evaluate import eval
from botiverse.models.TRIPPY.config import TRIPPYConfig
from botiverse.models.TRIPPY.TRIPPY import TRIPPY


class TRIPPYDST:
    """
    TRIPPYDST is a class that represents the TripPy Dialogue State Tracker.

    It provides methods for loading the model, training the model, updating the dialogue state,
    getting the current dialogue state, deleting slots, resetting the tracker, and displaying the tracker information.
    
    :param domains: The list of domains to consider.
    :type domains: list[str]
    :param slot_list: List of slot names.
    :type slot_list: list[str]
    :param label_maps: Dictionary of the variants of the slot-values that are mapped to the canonical slot-values.
    :type label_maps: dict[str, list]
    :param non_referable_slots: The list of non-referable slots.
    :type non_referable_slots: list[str]
    :param non_referable_pairs: The list of non-referable slot pairs.
    :type non_referable_pairs: list[tuple[str, str]]
    :param from_scratch: Whether to train the model from scratch.
    :type from_scratch: bool
    :param TRIPPY_config: The configuration for the TRIPPY model, defaults to TRIPPYConfig()
    :type TRIPPY_config: TRIPPYConfig, optional
    """

    def __init__(self, domains, slot_list, label_maps, non_referable_slots, non_referable_pairs, from_scratch, BERT_config, TRIPPY_config=TRIPPYConfig()):
        self.domains = domains
        self.non_referable_slots = non_referable_slots
        self.non_referable_pairs = non_referable_pairs
        self.from_scratch = from_scratch
        self.BERT_config = BERT_config
        self.TRIPPY_config = TRIPPY_config

        slot_list = fix_slot_list(slot_list, domains)
        self.slot_list = slot_list
        self.n_slots = len(slot_list)
        self.label_maps = label_maps

        self.state = {}
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.model = TRIPPY(len(slot_list), TRIPPY_config.hid_dim, TRIPPY_config.n_oper, TRIPPY_config.dropout, from_scratch, BERT_config, TRIPPY_config).to(self.device)
        self.history = []

    def save_model(self, model_path):
        """
        Save the trained model.

        :param model_path: The path to save the model.
        :type model_path: str
        """
        torch.save(self.model.state_dict(), model_path)

    def load_model(self, model_path, pretrained, test_path):
      """
      Load the trained model if path is given, else load a pretrained model.

      :param model_path: The path to the saved model.
      :type model_path: str
      :param test_path: The path to the test data for evaluation.
      :type test_path: str
      """

      # If model_path is None will the load a pretrained model.
      if model_path is None and pretrained == 'sim-R':
        # Download DST weights trained on sim-R
        curr_dir = os.path.dirname(os.path.abspath(__file__))
        model_path = os.path.join(curr_dir, 'sim-R.pt')
        
        f_id = '1POjBULmqxBrebvINfl989bskAstV3Zld'
        file_url = f'https://drive.google.com/uc?export=download&confirm=pbef&id={f_id}'
        if not os.path.exists(model_path):
            gdown.download(file_url, model_path)
            print('Model downloaded successfully.')
        else:
            print('Model already exists. Skipping download.')

      if self.from_scratch == True:
          # Get saved weights
          state_dict = torch.load(model_path, map_location=self.device)

          # Delete position_ids from the state_dict if available
          if 'bert.embeddings.position_ids' in state_dict.keys():
              del state_dict['bert.embeddings.position_ids']

          # Get the new weights keys from the model
          new_keys = list(self.model.state_dict().keys())

          # Get the weights from the state_dict
          old_keys = list(state_dict.keys())
          weights = list(state_dict.values())

          # Create a new state_dict with the new keys
          new_state_dict = OrderedDict()
          for i in range(len(new_keys)):
              new_state_dict[new_keys[i]] = weights[i]
              # print(old_keys[i], '->', new_keys[i])

          self.model.load_state_dict(new_state_dict)

      else:
          self.model.load_state_dict(torch.load(model_path, map_location=self.device))

      print('Model loaded successfully.')
      if test_path is not None:
        print('Preprocessing the data...')
        test_raw_data, test_data = prepare_data(test_path, self.slot_list, self.label_maps, self.TRIPPY_config.tokenizer, self.TRIPPY_config.max_len, self.domains, self.non_referable_slots, self.non_referable_pairs, self.TRIPPY_config.multiwoz)
        test_dataset = Dataset(test_data, self.n_slots, self.TRIPPY_config.oper2id, self.slot_list)
        test_data_loader = torch.utils.data.DataLoader(test_dataset,
                                                       batch_size=self.TRIPPY_config.test_batch_size)
        print('Evaluating the model on the data...')
        joint_goal_acc, per_slot_acc, macro_f1_score, all_f1_score = eval(test_raw_data, test_data, self.model, self.device, self.n_slots, self.slot_list, self.label_maps, self.TRIPPY_config.oper2id, self.TRIPPY_config.multiwoz)
        print(f'Joint Goal Acc: {joint_goal_acc}')
        print(f'Per Slot Acc: {per_slot_acc}')
        print(f'Macro F1 Score: {macro_f1_score}')
        print(f'All f1 score = {all_f1_score}')


    def train(self, train_path, dev_path=None, test_path=None):
      """
      Train the model.

      :param train_path: The path to the training data.
      :type train_path: str
      :param dev_path: The path to the development data for evaluation during training.
      :type dev_path: str
      :param test_path: The path to the test data for evaluation after training.
      :type test_path: str
      """

      # Save the temprorary model in the current directory.
      curr_dir = os.path.dirname(os.path.abspath(__file__))
      model_path = os.path.join(curr_dir, 'model.pt')

      # Train the model
      run(self.model, self.domains, self.slot_list, self.label_maps, train_path, dev_path, test_path, self.device, self.non_referable_slots, self.non_referable_pairs, model_path, self.TRIPPY_config)


    def update_state(self, sys_utter, user_utter, inform_mem):
      """
      Update the dialogue state based on the system and user utterances.

      :param sys_utter: The system utterance.
      :type sys_utter: str
      :param user_utter: The user utterance.
      :type user_utter: str
      :param inform_mem: The inform memory containing previous slot-value pairs.
      :type inform_mem: dict[str, list[str]]
      :return: The updated dialogue state.
      :rtype: dict[str, str]
      """

      # normalize utterances
      user_utter = ' '.join(normalize(user_utter, self.TRIPPY_config.multiwoz))
      sys_utter = ' '.join(normalize(sys_utter, self.TRIPPY_config.multiwoz))
      # delex the system utterance
      sys_utter = ' '.join(mask_utterance(sys_utter, inform_mem, self.TRIPPY_config.multiwoz, '[UNK]'))

      self.state = infer(self.model, self.slot_list, self.state, self.history, sys_utter, user_utter, inform_mem, self.device, self.TRIPPY_config.oper2id, self.TRIPPY_config.tokenizer, self.TRIPPY_config.max_len)
      self.history = [user_utter, sys_utter] + self.history
      return self.state.copy()

    def get_dialogue_state(self):
      """
      Get a copy of the current dialogue state.

      :return: A copy of the dialogue state.
      :rtype: dict[str, str]
      """
      return self.state.copy()

    def delete_slots(self, domain, slot):
      """
      Delete slots from the dialogue state.

      If a domain is specified, all slots in that domain will be deleted.
      If a slot is specified, that specific slot will be deleted.
      If neither domain nor slot is specified, all slots will be deleted.

      :param domain: The domain to delete slots from.
      :type domain: str
      :param slot: The slot to delete.
      :type slot: str
      """
      keys = self.state.keys()
      if domain is not None:
        for key in keys:
          if domain in key:
            del self.state[key]
      elif slot is not None:
          if slot in keys:
            del self.state[slot]
      else:
        for key in keys:
          del self.state[key]

    def reset(self):
      """
      Reset the dialogue state.

      Remove all slots from the dialogue state and clear the history.
      """
      keys = list(self.state.keys())
      for key in keys:
        del self.state[key]

      self.history = []


    def __str__(self):
      """
      Return a string representation of the TRIPPYDST object.

      :return: A string representation of the object.
      :rtype: str
      """
      string = ''
      string = string + '\ndomains: ' + str(self.domains)
      string = string + '\nslot_list: ' + str(self.slot_list)
      string = string + '\nlabel_maps: ' + str(self.label_maps)
      string = string + '\nnon_referable_slots: ' + str(self.non_referable_slots)
      string = string + '\nnon_referable_pairs: ' + str(self.non_referable_pairs)
      string = string + '\nfrom_scratch: ' + str(self.from_scratch)
      string = string + '\nslot_list: ' + str(self.slot_list)
      string = string + '\nn_slots: ' + str(self.n_slots)
      string = string + '\nlabel_maps: ' + str(self.label_maps)
      string = string + '\nstate: ' + str(self.state)
      string = string + '\ndevice: ' + str(self.device)
      # string = string + '\nmodel: ' + str(self.model)
      string = string + '\nhistory: ' + str(self.history)
      return string


#!/usr/bin/env python
# coding: utf-8

# # <font color="cyan">LSTM </font> From Scratch Implementation
# 
# In this notebook, we shall demonstrate implementing LSTM from scratch; up from linear layers.

# ### <font color='white'>LSTM Layer</font>
# 
# An LSTM layer takes in the input $x_t$ and the previous hidden and cell state $h_{t-1}$, $c_{t-1}$. It outputs the next hidden and cell state $h_t$, $c_t$ in the following fashion:
# 
# Define three gates and pass the input, previous hidden and cell state through them to get the outputs of the gates:
# 
# $$i_t = σ(W_{xi}x_t + W_{hi}h_{t-1} + b_i)$$
# 
# $$f_t = σ(W_{xf}x_t + W_{hf}h_{t-1} +  b_f)$$
# 
# $$o_t = σ(W_{xo}x_t + W_{ho}h_{t-1} + b_o)$$
# 
# Compute a candidate cell state and cell stage using another linear layer and input and forget gates:
# $$\hat{c} = tanh(W_{xg}x_t + W_{hg}h_{t-1} + b_g)$$
# 
# $$c_t = f_tc_{t-1} + i_tg_t$$
# 
# Compute the next hidden state using the computed cell state and output gate:
# $$h_t = o_ttanh(c_t)$$
# 
# 

# In[1]:


import torch
import torch.nn as nn
from torch import sigmoid, tanh as σ, tanh
from torch.autograd import Variable
import numpy as np
from tqdm import tqdm
import os

class LSTMCell(nn.Module): 
    '''
    An interface for a single LSTM layer.
    '''
    def __init__(self, input_size, hidden_size):
        '''
        Initialize the parameters single LSTM layer given the size of the inputs and the required hidden size.
        :param input_size: The size of the input to the LSTM layer
        :param hidden_size: The size of the hidden state of the LSTM layer
        '''
        super(LSTMCell, self).__init__()
        # input and output dimensions
        self.input_size = input_size
        self.hidden_size = hidden_size
        
        # LSTM involves three gates: input, forget, output and tanh linear layer.
        # We can declare them as follows and chunk them later since their two weights are of same size.
        self.Wₓₕ = nn.Linear(input_size, hidden_size * 4)
        self.Wₕₕ = nn.Linear(hidden_size, hidden_size * 4)
        
        # initalize the weights using Xavier initialization
        σ = 1.0 / np.sqrt(self.hidden_size)
        for w in self.parameters():     w.data.uniform_(-σ, σ)

    def forward(self, input, h, c):
        '''
        Forward pass of the LSTM layer which passes in the input and previous states and returns the new hidden and cell states
        :param input: The input to the LSTM layer which is of shape (batch_size, input_size)
        :param h: The previous hidden state of the LSTM layer which is of shape (batch_size, hidden_size)
        :param c: The previous cell state of the LSTM layer which is of shape (batch_size, hidden_size)
        :return: The new hidden state and cell state of the LSTM layer which are of shapes (batch_size, hidden_size) each
        '''
        # chunk the weights and biases of the gates
        weights = self.Wₓₕ(input) + self.Wₕₕ(h)
        Γi, Γf, T, Γo = weights.chunk(4, 1)

        # Get gates (iₜ, fₜ,  oₜ)
        iₜ, fₜ, oₜ = σ(Γi), σ(Γf), σ(Γo)
        
        # compute candidate and new cell state
        ĉ = tanh(T)
        cₜ  = c * fₜ + iₜ * ĉ

        # compute new hidden state
        hₜ = oₜ * tanh(cₜ)

        return (hₜ, cₜ)


# ### LSTM Class
# 
# Given an input sequence, each token passes by all the layers and each layer has its own hidden state and cell state which is its output due to the previous token.

# In[2]:


class LSTMX(nn.Module):
    '''
    An interface for a multi-layer LSTM where the hidden state of each layer is of the same size.
    '''
    def __init__(self, input_size, hidden_size, num_layers=1):
        '''
        Initialize the parameters of an LSTM with an arbitrary number of layers all of which have the same hidden size.
        :param input_size: The size of the input to the LSTM layer
        :param hidden_size: The size of the hidden state of the LSTM layer
        :param num_layers: The number of LSTM layers stacked on top of each other (default: 1)
        '''
        super(LSTMX, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers

        # initialize num_layers of LSTM layers
        self.lstm_layers = nn.ModuleList([LSTMCell(input_size, hidden_size)])
        for l in range(1, self.num_layers):
            self.lstm_layers.append(LSTMCell(self.hidden_size, self.hidden_size))


    def forward(self, input, hₒ=None):
        '''
        Forward pass of the LSTM which takes the input and initial hidden/cell state and returns the output
        due to the last token in the sequence.
        :param input: The input to the LSTM layer which is of shape (batch_size, seq_len, input_size)
        :param hₒ: The initial hidden and cell states of the LSTM layer which are of shape (num_layers, batch_size, hidden_size)
        :return: The output due to the last token in the sequence which is of shape (batch_size, hidden_size)
        '''
        # Input of shape (batch_size, seqence length , input_size)
        #
        # Output of shape (batch_size, output_size)
        batch_size, seq_len = input.size(0), input.size(1)
        
        if hₒ is None:
             v = Variable(input.new_zeros(self.num_layers, batch_size, self.hidden_size))
             hₒ = v.cuda() if torch.cuda.is_available() and hₒ else v
        
        # Will contain the output of the last layer for each token in the sequence
        outs = []

        H = [hₒ[l, :, :] for l in range(self.num_layers)]
        C = [hₒ[l, :, :] for l in range(self.num_layers)]
        # Now H[l], C[l] are the current hidden state for layer l (defined for every sequence in the batch)
        
        # for each token in the sequence
        for t in range(seq_len):
            # pass it through each layer
            token = input[:, t, :]
            for l in range(self.num_layers):
                lstm = self.lstm_layers[l]   # takes input (batch_size, input_size) and h, c of shape (batch_size, hidden_size)
                # layer takes the token or output from the previous layer and initial hidden, cell states
                H[l], C[l] = lstm(token, H[l], C[l]) if l == 0 else lstm(H[l - 1], H[l], C[l])

            # the output due to any token is that due to the last layer of the lstm
            outs.append(H[l])

        return outs[-1]


# In[3]:


class LSTMClassifier(nn.Module):
    '''
    LSTMClassifier class defines a high-level interface of the LSTM model for classification.
    '''
    def __init__(self, input_size, hidden_size, num_classes):
        super(LSTMClassifier, self).__init__()
        self.hidden_size = hidden_size
        self.lstm = LSTMX(input_size, hidden_size)
        self.fc = nn.Linear(hidden_size, num_classes)
        self.criterion = nn.CrossEntropyLoss()

    def forward(self, x):
        '''
        Forward pass of the LSTMClassifier which takes the input and passes it through all the LSTM layers and an output layer to produce an output.
        :param x: The input to the LSTMClassifier which is of shape (batch_size, seq_len, input_size)
        :return: The output of the LSTMClassifier which is of shape (batch_size, num_classes)
        '''
        out = self.lstm(x)
        out = self.fc(out)
        return out
    
    
    def fit(self, X, y, λ=0.001, α=1e-3, max_epochs=100, patience=5, val_ratio=0.2):
        '''
        Fit the LSTMClassifier to the given data.
        :param X: The input data of shape (batch_size, seq_len, input_size)
        :param y: The labels of the data of shape (batch_size)
        :param hidden_size: The size of the hidden state of the LSTM layer (default: 64)
        :param λ: The learning rate (default: 0.001)
        :param num_epochs: The number of epochs to train the model for (default: 100)
        '''
        Xt = torch.from_numpy(X)
        yt = torch.from_numpy(y)
        if val_ratio:
            indices = torch.randperm(len(Xt))
            Xt, yt = Xt[indices], yt[indices]
            # split the data into train and validation sets
            val_size = int(val_ratio * len(Xt))
            Xt, Xv = Xt[:-val_size], Xt[-val_size:]
            yt, yv = yt[:-val_size], yt[-val_size:]
            self.Xv, self.yv = Xv, yv
                
        optimizer = torch.optim.Adam(self.parameters(), lr=λ, weight_decay=α)
        print("Training the LSTMClassifier...")
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        bad_epochs = 0
        val_accuracy = 0
        val_loss = 0
        best_loss = np.inf
        pbar = tqdm(range(max_epochs))
        for epoch in pbar:
            outputs = self(Xt)
            loss = self.criterion(outputs.squeeze(), yt)
            pbar.set_description(f"Epoch {epoch+1}/{max_epochs}, Loss: {loss.item()}")
            
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            if val_ratio:
                # randomly shuffle the data
                val_accuracy = self.evaluate(Xv, yv)
                with torch.no_grad():
                    val_loss = self.criterion(self(Xv).squeeze(), yv)
                if val_loss < best_loss:
                    best_loss = val_loss
                    bad_epochs = 0
                    # save the model
                    torch.save(self.state_dict(), os.path.join(curr_dir, "LSTMClassifier.pt"))
                else:
                    bad_epochs += 1
                    if bad_epochs == patience:
                        print(f"{patience} epochs have passed without improvement. Early stopping...")
                        self.load_state_dict(torch.load(os.path.join(curr_dir, "LSTMClassifier.pt")))
                        break
                # every 5 epochs see
                pbar.set_postfix({"Validation Accuracy": val_accuracy})             
           

    def predict(self, X):
        '''
        Predict the labels of the given data by passing it through the LSTMClassifier.
        :param X: The input data of shape (batch_size, seq_len, input_size)
        :return: The predicted labels of the data of shape (batch_size)
        '''
        Xt = torch.from_numpy(X)
        outputs = self(Xt)
        pred = torch.argmax(outputs, dim=1)
        softmax = nn.Softmax(dim=1)
        prob = torch.max(softmax(outputs), dim=1)
        return pred.detach().numpy(), prob.values.detach().numpy()
    
    def evaluate(self, Xt, yt):
        '''
        Evaluate the LSTMClassifier on the given data.
        :param X: The input data of shape (batch_size, seq_len, input_size)
        :param y: The labels of the data of shape (batch_size)
        :return: The accuracy of the LSTMClassifier on the given data
        '''
        # check ig they are torch tensors
        if not isinstance(Xt, torch.Tensor) or not isinstance(yt, torch.Tensor):
            Xt = torch.from_numpy(Xt)
            yt = torch.from_numpy(yt)
        outputs = self(Xt)
        outputs = torch.argmax(outputs, dim=1)
        # compute the accuracy
        return (outputs == yt).sum().item() / len(yt)
    
    def save(self, path):
        '''
        Save the LSTMClassifier to a file.
        :param path: The path to the file
        '''
        torch.save(self.state_dict(), path)
    
    def load(self, path):
        '''
        Load the LSTMClassifier from a file.
        :param path: The path to the file
        '''
        self.load_state_dict(torch.load(path))


# In[4]:


# if running from notebook
if __name__ == '__main__':
    get_ipython().system('jupyter nbconvert --to script LSTM.ipynb')




"""
Theis Module contains the BERT model architecture.
"""

import torch
import torch.nn as nn



# Embeddings
# 1. Cluster similar words together.
# 2. Preserve different relationships between words such as: semantic, syntactic, linear,
# and since BERT is bidirectional it will also preserve contextual relationships as well.
class Embeddings(nn.Module):
    """
    Embedding layer for BERT.

    This layer takes input_ids and token_type_ids as inputs and generates word embeddings
    using three types of embeddings: word, position, and token_type embeddings.

    :param config: BERT configuration.
    :type config: Config
    """
    def __init__(self, config):
        super(Embeddings, self).__init__()
        # Bert uses 3 types of embeddings: word, position, and token_type (segment type).
        # LayerNorm is used to normalize the sum of the embeddings.
        self.word_embeddings = nn.Embedding(config.vocab_size, config.hidden_size, padding_idx=config.padding_idx)
        self.position_embeddings = nn.Embedding(config.max_seq, config.hidden_size)
        self.token_type_embeddings = nn.Embedding(config.token_types, config.hidden_size)
        self.layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, input_ids, token_type_ids): # input_ids: [batch_size, seq_len] token_type_ids: [batch_size, seq_len]
        """
        Forward pass of the Embeddings layer.

        :param input_ids: The input token IDs.
        :type input_ids: torch.Tensor
        :param token_type_ids: The token type IDs.
        :type token_type_ids: torch.Tensor
        :return: The generated embeddings.
        :rtype: torch.Tensor
        """
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        seq_len = input_ids.size(1)
        position_ids = torch.arange(seq_len).unsqueeze(0).expand_as(input_ids).to(device) # position_ids: [batch_size, seq_len]
        word_embeddings = self.word_embeddings(input_ids)  # word_embeddings: [batch_size, seq_len, hidden_size]
        position_embeddings = self.position_embeddings(position_ids) # position_embeddings: [batch_size, seq_len, hidden_size]
        token_type_embeddings = self.token_type_embeddings(token_type_ids) # token_type_embeddings: [batch_size, seq_len, hidden_size]
        embeddings = word_embeddings + position_embeddings + token_type_embeddings # embeddings: [batch_size, seq_len, hidden_size]
        # Normalize by subtracting the mean and dividing by the standard deviation calculated across the feature dimension
        # then multiply by a learned gain parameter and add to a learned bias parameter.
        embeddings = self.layer_norm(embeddings) # embeddings: [batch_size, seq_len, hidden_size]
        embeddings = self.dropout(embeddings) # embeddings: [batch_size, seq_len, hidden_size]
        return embeddings

# Encoder layer
class EncoderLayer(nn.Module):
    """
    Encoder layer for BERT.

    This layer contains self-attention, layer normalization, and position-wise feed-forward network.

    :param config: BERT configuration.
    :type config: Config
    """

    def __init__(self, config):
        super(EncoderLayer, self).__init__()
        self.self_attention = MultiHeadAttention(config)
        self.self_layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.self_dropout = nn.Dropout(config.dropout)
        self.position_wise_feed_forward = PositionWiseFeedForward(config)
        self.ffn_layer_norm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.ffn_dropout = nn.Dropout(config.dropout)

    def forward(self, input, attention_mask):
        """
        Forward pass of the EncoderLayer.

        :param input: The input tensor.
        :type input: torch.Tensor
        :param attention_mask: The attention mask.
        :type attention_mask: torch.Tensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """

        # Multi-head attention
        context, attention = self.self_attention(input, input, input, attention_mask) # context: [batch_size, seq_len, hidden_size] attention: [batch_size, heads, seq_len, seq_len]
        # Add and normalize
        context = self.self_dropout(context) # context: [batch_size, seq_len, hidden_size]
        output = self.self_layer_norm(input + context) # output: [batch_size, seq_len, hidden_size]
        # Position-wise feed-forward network
        context = self.position_wise_feed_forward(output) # context: [batch_size, seq_len, hidden_size]
        # Add and normalize
        context = self.ffn_dropout(context) # context: [batch_size, seq_len, hidden_size]
        output = self.ffn_layer_norm(output + context) # output: [batch_size, seq_len, hidden_size]
        return output, attention

# Multi-head attention
class MultiHeadAttention(nn.Module):
    """
    Multi-head attention layer for BERT.

    This layer performs multi-head self-attention and returns the output context.

    :param config: BERT configuration.
    :type config: Config
    """
    def __init__(self, config):
        super(MultiHeadAttention, self).__init__()
        self.config = config
        self.w_q = nn.Linear(config.hidden_size, config.hidden_size)
        self.w_k = nn.Linear(config.hidden_size, config.hidden_size)
        self.w_v = nn.Linear(config.hidden_size, config.hidden_size)
        self.w_o = nn.Linear(config.hidden_size, config.hidden_size)
        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, query, key, value, attention_mask): 
        """
        Forward pass of the MultiHeadAttention.

        :param query: The query tensor.
        :type query: torch.Tensor
        :param key: The key tensor.
        :type key: torch.Tensor
        :param value: The value tensor.
        :type value: torch.Tensor
        :param attention_mask: The attention mask.
        :type attention_mask: torch.Tensor
        :return: The output context.
        :rtype: torch.Tensor
        """                                                  
                
        # query: [batch_size, seq_len, hidden_size] key: [batch_size, seq_len, hidden_size]
        # value: [batch_size, seq_len, hidden_size] attention_mask: [batch_size, seq_len_q, seq_len_k]                                          

        batch_size, seq_len, hidden_size = query.size()

        query = self.w_q(query).view(batch_size, seq_len, self.config.heads, hidden_size // self.config.heads).transpose(1, 2) # query: [batch_size, heads, seq_len, hidden_size // heads]
        key = self.w_k(key).view(batch_size, seq_len, self.config.heads, hidden_size // self.config.heads).transpose(1, 2) # key: [batch_size, heads, seq_len, hidden_size // heads]
        value = self.w_v(value).view(batch_size, seq_len, self.config.heads, hidden_size // self.config.heads).transpose(1, 2) # value: [batch_size, heads, seq_len, hidden_size // heads]

        # Scaled dot-product attention
        attention = torch.matmul(query, key.transpose(-1, -2)) / torch.sqrt(torch.tensor(float(hidden_size // self.config.heads))) # attention: [batch_size, heads, seq_len, seq_len]
        attention_mask = attention_mask.unsqueeze(1).repeat(1, self.config.heads, 1, 1) # attention_mask: [batch_size, heads, seq_len_q, seq_len_k]
        attention_mask = (attention_mask == 0)
        attention.masked_fill_(attention_mask, -1e9) # attention: [batch_size, heads, seq_len, seq_len]
        attention = self.softmax(attention) # attention: [batch_size, heads, seq_len, seq_len]
        attention = self.dropout(attention) # attention: [batch_size, heads, seq_len, seq_len]
        context = torch.matmul(attention, value) # context: [batch_size, heads, seq_len, hidden_size // heads]
        context = context.transpose(1, 2).contiguous().view(batch_size, seq_len, hidden_size) # context: [batch_size, seq_len, hidden_size]
        output = self.w_o(context) # output: [batch_size, seq_len, hidden_size]
        return output, attention

# Position-wise feed-forward network
class PositionWiseFeedForward(nn.Module):
    """
    Position-wise feed-forward network layer for BERT.

    This layer applies two linear transformations with a GELU activation function.

    :param config: BERT configuration.
    :type config: Config
    """
    def __init__(self, config):
        super(PositionWiseFeedForward, self).__init__()
        self.linear1 = nn.Linear(config.hidden_size, config.ff_size)
        self.linear2 = nn.Linear(config.ff_size, config.hidden_size)
        self.gelu = nn.GELU()

    def forward(self, input):
        """
        Forward pass of the PositionWiseFeedForward layer.

        :param input: The input tensor.
        :type input: torch.Tensor
        :return: The output tensor.
        :rtype: torch.Tensor
        """
        output = self.linear1(input) # output: [batch_size, seq_len, ff_size]
        output = self.gelu(output) # output: [batch_size, seq_len, ff_size]
        output = self.linear2(output) # output: [batch_size, seq_len, hidden_size]
        return output

# Bert
# 1. Puts it all together.
class Bert(nn.Module):
    """
    BERT model implementation.

    This model combines the Embeddings layer, EncoderLayers, and linear transformation layers
    to perform BERT-based processing.

    :param config: BERT configuration.
    :type config: Config
    """
    def __init__(self, config):
        super(Bert, self).__init__()
        self.embeddings = Embeddings(config)
        self.encoder = nn.ModuleList([EncoderLayer(config) for _ in range(config.encoder_layers)])
        self.linear = nn.Linear(config.hidden_size, config.hidden_size)
        self.tanh = nn.Tanh()

    def forward(self, input_ids, token_type_ids, attention_mask, return_dict=False): # input_ids: [batch_size, seq_len] token_type_ids: [batch_size, seq_len] attention_mask: [batch_size, seq_len]
        """
        Forward pass of the Bert model.

        :param input_ids: The input token IDs.
        :type input_ids: torch.Tensor
        :param token_type_ids: The token type IDs.
        :type token_type_ids: torch.Tensor
        :param attention_mask: The attention mask.
        :type attention_mask: torch.Tensor
        :param return_dict: Whether to return a dictionary or not, defaults to False.
        :type return_dict: bool
        :return: The sequence output and pooled output.
        :rtype: torch.Tensor, torch.Tensor
        """
        
        # Embedding
        output = self.embeddings(input_ids, token_type_ids) # output: [batch_size, seq_len, hidden_size]

        # Encoder
        attention_mask = attention_mask.unsqueeze(1).repeat(1, output.size(1), 1) # attention_mask: [batch_size, seq_len, seq_len]
        for encoder_layer in self.encoder:
            output, attention = encoder_layer(output, attention_mask) # output: [batch_size, seq_len, hidden_size] attention: [batch_size, heads, seq_len, seq_len]

        # Sequnce and pooled outputs
        sequence_output = output # sequence_output: [batch_size, seq_len, hidden_size]
        pooled_output = self.tanh(self.linear(sequence_output[:, 0])) # pooled_output: [batch_size, hidden_size]

        return sequence_output, pooled_output


import torch
import torch.nn as nn
import torch.optim as optim
from tqdm.auto import tqdm
from botiverse.models.GRUClassifier.GRUClassifier import GRUTextClassifier
from botiverse.preprocessors.Special.WhizBot_GRU_Preprocessor.WhizBot_GRU_Preprocessor import WhizBot_GRU_Preprocessor
import random

class WhizBot_GRU:
    '''An interface for the WhizBot_GRU model which is based on simple GRU model'''
    def __init__(self):
        """
        Initializes WhizBot_GRU, and will prepare the GPU device based on CUDA availability.
        """
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def read_data(self, file_path):
        """
        Reads and pre-processes the data, sets up the model based on the data and prepares the train-validation split.

        :param file_path: The path to the file that contains the dataset.
        :type file_path: str

        :returns: None
        """
        # read data
        self.preprocessor = WhizBot_GRU_Preprocessor(file_path)
        # process data
        self.data = self.preprocessor.process()
        # prepare model
        vocab_size = len(self.preprocessor.tokenizer.get_vocab())
        num_labels = len(self.preprocessor.label_dict)
        self.model = GRUTextClassifier(vocab_size, 300, num_labels).to(self.device)
        self.criterion = nn.CrossEntropyLoss()
        self.optimizer = optim.SGD(self.model.parameters(), lr=0.05)
        # train validation split
        self.train_data = self.data.sample(frac=0.8, random_state=0)
        self.validation_data = self.data.drop(self.train_data.index)
        self.train_data = self.train_data.reset_index(drop=True)
        self.validation_data = self.validation_data.reset_index(drop=True)

    def train(self, epochs=50, batch_size=32):
        """
        Trains the model using the training dataset.

        :param epochs: The number of training epochs.
        :type epochs: int

        :param batch_size: The number of training examples utilized used to make one paramenters updat.
        :type batch_size: int

        :returns: None
        """
        self.model.train()
        for epoch in range(epochs):
            for i in tqdm(range(0, len(self.train_data), batch_size)):
                self.model.zero_grad()
                # get the batches
                batch_texts = torch.stack(self.train_data['text'][i:i+batch_size].tolist()).to(self.device)
                batch_labels = torch.cat(self.train_data['label'][i:i+batch_size].tolist()).to(self.device)
                output = self.model(batch_texts)
                loss = self.criterion(output, batch_labels)
                # backpropagate
                loss.backward()
                self.optimizer.step()
            print("Epoch: " + str(epoch) + " Loss: " + str(loss.item()))

    def validation(self, batch_size=32):
        """
        Tests the model performance using the validation dataset and calculates the accuracy.
        
        :param batch_size: The number of training examples utilized used to make one paramenters updat.
        :type batch_size: int

        :returns: None
        """
        correct = 0
        total = 0
        self.model.eval()
        with torch.no_grad():
            for i in tqdm(range(0, len(self.validation_data), batch_size)):
                # get the batches
                batch_texts = torch.stack(self.validation_data['text'][i:i+batch_size].tolist()).to(self.device)
                batch_labels = torch.cat(self.validation_data['label'][i:i+batch_size].tolist()).to(self.device)
                outputs = self.model(batch_texts)
                _, predicted = torch.max(outputs.data, 1)
                # calculate accuracy
                total += batch_labels.size(0)
                correct += (predicted == batch_labels).sum().item()
        print('Accuracy: %d %%' % (100 * correct / total))

    def infer(self, string):
        """
        Performs inference using the model.

        :param string: The input string to perform inference on.
        :type string: str

        :returns: A random response from the response list of the predicted label.
        """
        self.model.eval()
        with torch.no_grad():
            string = self.preprocessor.process_string(string)
            string = string.unsqueeze(0).to(self.device)
            output = self.model(string)
            _, predicted = torch.max(output.data, 1)
            # get the label
            for key, value in self.preprocessor.label_dict.items():
                if value == predicted.item():
                    label = key
                    break
        # return a random responce of the label
        return random.choice(self.preprocessor.responces[label])

    def save(self, path):
        """
        Saves the model parameters to the given path.

        :param path: The path where the model parameters will be saved.
        :type path: str

        :returns: None
        """
        torch.save(self.model.state_dict(), path)

    def load(self, path):
        """
        Loads the model parameters from the given path.

        :param path: The path from where the model parameters will be loaded.
        :type path: str

        :returns: None
        """
        self.model.load_state_dict(torch.load(path))


#!/usr/bin/env python
# coding: utf-8

# # <font color="cyan">FastSpeech 1.0 </font> Implementation
# 
# In this notebook, we shall demonstrate implementing FastSpeech 1.0 from scratch for inference purposes. 

# In[1]:


'''
FastSpeech 1.0 interface and implementation from scratch in PyTorch for inference.
'''
import torch
import torch.nn as nn
from torch.nn import functional as F
import numpy as np

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


# 
# ![Image](https://i.imgur.com/ZDR7wqr.png)
# 

# #### Clearly, we have three main components in the model:
# 
# <div align='center'>
# <table>
#   <tr>
#     <th colspan="1"><font color="yellow">Component</font></th>
#     <th colspan="3"><font color="yellow">Encoder</font></th>
#     <th colspan="2"><font color="yellow">Length Regulator</font></th>
#     <th colspan="3"><font color="yellow">Decoder</font></th>
# 
#   </tr>
#   <tr>
#     <th colspan="1"><font color="white">Subcomponenets</font></th>
#     <th colspan="1"><font color="white">Phoneme Embedding</font></th>
#     <th colspan="1"><font color="white">Positional Encoding</font></th>
#     <th colspan="1"><font color="white">FFT Block</font></th>
#     <th colspan="1"><font color="white">Duration Predictor</font></th>
#     <th colspan="1"><font color="white">LR Logic</font></th>
#     <th colspan="1"><font color="white">Positional Encoding</font></th>
#     <th colspan="1"><font color="white">FFT Block</font></th>
#     <th colspan="1"><font color="white">Linear Layer</font></th>
#   </tr>
# 
#   <tr>
#   <th colspan="1">Takes</th>
#   <td colspan="3"> <font color='cyan'>[batch_size, seq_len]</font></td>
#   <td colspan="2"> <font color='cyan'>[batch_size, seq_len, emb_dim]</font> </td>
#   <td colspan="3"> <font color='cyan'>[batch_size, new_seq_len, emb_dim]</font> </td>
#   </tr>
# 
#   <tr>
#   <th colspan="1">Yields</th>
#   <td colspan="3"> <font color='cyan'>[batch_size, seq_len, emb_dim]  </td>
#   <td colspan="2"> <font color='cyan'>[batch_size, new_seq_len, emb_dim]</font> </td>
#   <td colspan="3"> <font color='cyan'>[batch_size, new_seq_len, mel_num]</font> </td>
#   </tr>
# 
#   <tr>
#   <th colspan="1">Purpose</th>
#   <td colspan="3">Learn a representation for phonemes. The three layers within guarantee that underlying words, how they are ordered and other words all take part in the representation. </td>
#   <td colspan="2">Predict the duration of each phoneme and repeat accordingly</td>
#   <td colspan="3">Given time-aligned phoneme representations learn to transform them into a mel spectogram</td>
#   </tr>
# </table>
# </div>
# 
# Once we have the spectogram, all that's needed it a vocoder to transform to audio by performing an approximate inverse mel-transform such as Griffin-Lim algorithm or using a pretrained model that does the task more accurately such as WaveGlow; hence, we will go with the latter but we won't implement WaveGlow from scratch.
# 

# #### We will start by implementing 
# 
# <div align='left'>
# <table>
#   <tr>
#     <th colspan="2"><font color="cyan">FFT Block</th>
#   </tr>
#   <tr>
#     <th colspan="1">Multi-head Attention</th>
#     <th colspan="1">Conv1DNet</th>
#   </tr>
# </table>
# </div>
# 
# #### Then we have that
# 
# <div align='left'>
# <table>
#   <tr>
#     <th colspan="3"><font color="deepskyblue">Encoder</font></th>
#     <th colspan="3"><font color="deepskyblue">Decoder</font></th>
#   </tr>
#   <tr>
#     <th colspan="1">Phoneme Emb.</th>
#     <th colspan="1">Positional Enc.</th>
#     <th colspan="1">FFT Block</th>
#     <th colspan="1">Positional Encoding</th>
#     <th colspan="1">FFT Block</th>
#     <th colspan="1">Linear Layer</th>
#   </tr>
# </table>
# </div>
# 
# #### So we follow with
# 
# <div align='left'>
# <table>
#   <tr>
#     <th colspan="2"><font color="deepskyblue">Length Regulator</font></th>
#   </tr>
#   <tr>
#     <th colspan="1">LR Logic</th>
#     <th colspan="1">Duration Predictor</th>
#   </tr>
# 
# </table>
# </div>
# 
# #### And we finally have that
# 
# <div align='left'>
# <table>
#   <tr>
#     <th colspan="3"><font color="deepskyblue">Model</font></th>
#   </tr>
#   <tr>
#     <th colspan="1">Encoder</th>
#     <th colspan="1">Length Regulator</th>
#     <th colspan="1">Decoder</th>
#   </tr>
# </table>
# </div>

# ### <font color="cyan">1. Feedforward-Transformer Block </font>

# ![img](https://i.imgur.com/Kb88BwT.png)

# #### <font color="white"> Multi-head Self-Attention </font>

# <div align='center'>
# 
# $MultiHead(Q, K, V) = \text{Concat}(head_1, \ldots, head_h) \cdot W^O$
# 
# where
# 
#   $head = \text{Attention}(qW_q, kW_k, vW_v) =\text{Attention}(Q, K, V)$
# 
#   such that
# 
#   $\text{Attention}(Q, K, V) = \text{softmax}\left(\frac{QK^T}{\sqrt{d_k}}\right) \cdot V$
# 
#   
#   </div>

# In[2]:


class MultiHeadAttention(nn.Module):
    '''
    Multi-Head Attention module with a residual connection and layer normalization. Used as self-attention in the FFT block of the encoder and decoder.
    '''
    def __init__(self, num_head, emb_dim, h_dim, dropout=0.1):
        '''
        Initialize the parameters and structure of the Multi-Head Attention module.
        num_head: number of heads
        emb_dim: input encoder/decoder embeddings dimensions
        h_dim: hidden dimension (output dimension of the linear layers Wq, Wk, Wv)
        dropout: dropout probability
        '''
        super().__init__()

        self.num_head = num_head
        self.h_dim = h_dim                            # dimensionality of the final ouput
        self.head_dim = h_dim // num_head             # dimensionality of each head

        self.Wq = nn.Linear(emb_dim, h_dim)         # Equivalent to using head_dim, num_head times
        self.Wk = nn.Linear(emb_dim, h_dim)
        self.Wv = nn.Linear(emb_dim, h_dim)

        self.softmax = nn.Softmax(dim=2)
        self.scale = torch.sqrt(torch.FloatTensor([self.head_dim])).to(device)
        
        self.layer_norm = nn.LayerNorm(emb_dim)
        self.dropout = nn.Dropout(dropout)

        self.fc = nn.Linear(h_dim, emb_dim)
        
       
    def forward(self, q, k, v, mask):
        '''
        Pass given query, key and value through the Multi-Head Attention module.
        q: query of shape [batch_size, seq_len, emb_dim]
        k: key of shape [batch_size, seq_len, emb_dim]
        v: value of shape [batch_size, seq_len, emb_dim]
        mask: mask to apply to the attention so that padding tokens do not attend and are not attended to.
        returns: output of shape [batch_size, seq_len, emb_dim]    
        '''
        residual = q
        batch_size, num_head, seq_len, head_dim = q.size(0), self.num_head, q.size(1), self.head_dim
        Q, K, V = self.Wq(q), self.Wk(k), self.Wv(v)              # [batch_size, seq_len, emb_dim -> h_dim]

        review = lambda X: X.view(batch_size, seq_len, num_head, head_dim).transpose(1, 2).contiguous()
        Q, K, V = review(Q), review(K), review(V)                  # [batch_size, num_head, seq_len, head_dim]
        
        reshape = lambda X: X.view(batch_size * num_head, seq_len, head_dim)
        Q, K, V = reshape(Q), reshape(K), reshape(V)               # [batch_size * num_head, seq_len, head_dim]

        mask = mask.repeat(num_head, 1, 1)                         # [batch_size * num_head, seq_len, seq_len]
        
        a = torch.bmm(Q, K.transpose(1, 2))/self.scale             # [batch_size * num_head, seq_len, seq_len]
        
        a = a.masked_fill(mask, -np.inf)

        a = self.softmax(a)
        
        a = self.dropout(a)
        
        output = torch.bmm(a, V)                                  # [batch_size * num_head, seq_len, head_dim]   
        
        output = output.view(batch_size, num_head, seq_len, head_dim).transpose(1, 2).contiguous()
        output = output.view(batch_size, seq_len, num_head * head_dim)
        # [batch_size, seq_len, num_head * head_dim]

        output = self.dropout(self.fc(output))
        output = self.layer_norm(output + residual)

        return output


# #### <font color="white"> Conv1D Network </font>

# $Embeddings ⇒ Conv1D ⇒ ReuLU ⇒ Conv1D ⇒  Dropout ⇒ AddNorm$

# In[3]:


class Conv1DNet(nn.Module):
    '''
    1D convolutional network with residual connection and layer normalization. Used in the FFT block of the encoder and decoder.
    '''
    
    def __init__(self, inp_dim, inner_dim, dropout=0.1):
        '''
        Initialize the parameters and structure of the 1D convolutional network.
        inp_dim: input dimension
        inner_dim: inner dimension of the convolutional layers (output dimension of the first convolutional layer)
        dropout: dropout probability
        '''
        super().__init__()
        # Both convolutions have SAME padding so they only change the number of 1D channels
        self.conv1 = nn.Conv1d(inp_dim, inner_dim, kernel_size=9, padding=4) # from n to n-9+2*4+1 = n
        self.relu = nn.ReLU()
        self.conv2 = nn.Conv1d(inner_dim, inp_dim, kernel_size=1, padding=0) # from n to n-1+2*0+1 = n
        self.layer_norm = nn.LayerNorm(inp_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):   
        '''
        Pass given input through the 1D convolutional network. The input comes from the Multi-Head Attention module.
        x: input of shape [batch_size, seq_len, d_in]
        returns: output of shape [batch_size, seq_len, d_in]
        '''
        residual = x                                     # [batch_size, seq_len, d_in]
        x = x.transpose(1, 2)                            # [batch_size, d_in, seq_len]
        x = F.relu(self.conv1(x))                        # [batch_size, d_out, seq_len]
        x = self.conv2(x)                                # [batch_size, din, seq_len]
        x = x.transpose(1, 2)                            # [batch_size, seq_len, din]
        
        output = self.dropout(x)
        output = self.layer_norm(output + residual)

        return output


# #### <font color="white"> FFT Block </font>
# 
# Here we just put the two pieces together

# In[4]:


class FFTBlock(torch.nn.Module):
    '''
    FFT block used in the encoder and decoder. It consists of a Multi-Head Attention module and a 1D convolutional network.
    '''
    def __init__(self,  emb_dim, num_head, h_dim,  inner_dim, dropout=0.1):
        '''
        Initialize the structure of the FFT block.
        emb_dim: input encoder/decoder embeddings dimensions
        inner_dim: inner dimension of the convolutional layers (output dimension of the first convolutional layer)
        num_head: number of heads for the Multi-Head Attention module
        h_dim: hidden dimension (output dimension of the linear layers Wq, Wk, Wv)
        dropout: dropout probability
        '''
        super(FFTBlock, self).__init__()
        self.attn_out = MultiHeadAttention(num_head, emb_dim, h_dim, dropout=dropout)
        self.conv_out = Conv1DNet(emb_dim, inner_dim, dropout=dropout)

    def forward(self, input, non_pad_mask, attn_mask):
        '''
        Pass given encoder/decoder embeddings through the FFT block. The input comes from the previous FFT block or the input embeddings.
        input: input of shape [batch_size, seq_len, emb_dim]
        non_pad_mask: mask to nullify outputs due to padding tokens.
        attn_mask: mask to apply to the attention so that future tokens do not attend and are not attended to.
        returns: output of shape [batch_size, seq_len, emb_dim] and attention weights of shape [batch_size * num_head, seq_len, seq_len]
        '''
        non_pad_mask = non_pad_mask.float()
        # From [batch_size, seq_len, emb_dim] to [batch_size, seq_len, h_dim->emb_dim]
        output = self.attn_out(input, input, input, attn_mask) * non_pad_mask

        # From [batch_size, seq_len, emb_dim] to [batch_size, seq_len, emb_dim]
        output = self.conv_out(output) * non_pad_mask

        return output


# ### <font color="cyan">2. Encoder</font> & <font color="cyan">Decoder</font> 

# ![](https://i.imgur.com/HKBCPO7.png)

# It's obvious that they share most of the structure except for the first/last layer. Hence, we will implement both in one module. But before we do we need positional encoding.

# #### <font color="white"> Positional Encoding </font>

# 
# $$PE_{(pos, 2i)} = \sin(\frac{{pos}}{{10000^{2i/d_{\text{inp}}}}})$$
# 
# 
# $$PE_{(pos, 2i+1)} = \cos(\frac{{pos}}{{10000^{2i/d_{\text{inp}}}}})$$
# 

# The purpose of this is to add some notion of order to the input sequence without distorting it by adding a positional depending signal

# In[5]:


class SinusoidEncodingTable:
    '''
    Sinusoid encoding table used in the encoder and decoder. It is used to add positional information to the input embeddings.
    '''
    def __init__(self, max_seq_len, inp_dim, padding_idx=None):
        '''
        Initialize the parameters and structure of the sinusoid encoding table.
        max_seq_len: maximum sequence length
        inp_dim: input encoder/decoder embeddings dimensions
        padding_idx: index of the padding token
        '''
        self.max_seq_len = max_seq_len + 1
        self.inp_dim = inp_dim
        self.padding_idx = padding_idx
        self.sinusoid_table = self.build_table()

    def build_table(self):
        '''
        Build the sinusoid encoding table.
        :returns: sinusoid encoding table of shape [max_seq_len, inp_dim] which is indexed by the position of the input embeddings and the index of input emeddings value.
        '''
        # compute the denominator
        inds = np.arange(self.inp_dim) // 2
        div_term = np.power(10000, 2 * inds / self.inp_dim)

        # compute the numerator
        positions = np.arange(self.max_seq_len)
                
        # compute table of shape [max_seq_len, inp_dim] 
        sinusoid_table = np.outer(positions, 1/div_term)

        # apply sin to even indices in the array; 2i and cos to odd indices; 2i+1
        sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2]) 
        sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  

        if self.padding_idx is not None:
            sinusoid_table[self.padding_idx] = 0.

        return torch.Tensor(sinusoid_table)


# #### <font color="white"> Dencoder </font>
# 
# This is both the encoder and decoder in one module

# In[6]:


class Dencoder(nn.Module):
    '''
    A single module that implements the encoder and decoder of the FastSpeech 1.0 model. 
    '''
    def __init__(self, mode, vocab_dim, max_seq_len, emb_dim, num_layer, num_head, h_dim, d_inner, mel_num, dropout):
        '''
        Initialize the parameters and structure of the encoder/decoder.
        mode: 'Encoder' or 'Decoder'
        vocab_dim: vocabulary dimension. None if mode is 'Decoder'
        max_seq_len: maximum sequence length as needed by the sinusoid encoding table
        emb_dim: encoder/decoder embeddings dimensions
        num_layer: number of FFT blocks
        num_head: number of heads for the Multi-Head Attention module
        h_dim: hidden dimension (output dimension of the linear layers Wq, Wk, Wv)
        d_inner: inner dimension of the convolutional layers (output dimension of the first convolutional layer)
        mel_num: number of mel spectrogram bins (to map the final decoder embeddings). None if mode is 'Encoder'
        dropout: dropout probability
        '''
        super(Dencoder, self).__init__()
        self.mode= mode

        self.embedding = nn.Embedding(vocab_dim, emb_dim, padding_idx=0) if self.mode == 'Encoder' else lambda x: x

        table = SinusoidEncodingTable(max_seq_len, emb_dim, padding_idx=0).sinusoid_table
        self.position_encoding = nn.Embedding.from_pretrained(table, freeze=True)

        self.layer_stack = nn.ModuleList([FFTBlock(emb_dim, num_head, h_dim, d_inner, dropout=dropout) for _ in range(num_layer)])

        self.linear = nn.Linear(emb_dim, mel_num) if self.mode == 'Decoder' else lambda x: x
            
    def forward(self, inp_seq, inp_seq_pos):                                            
        '''
        Pass given input sequence through the encoder/decoder.
        inp_seq: input sequence of shape [batch_size, seq_len] for Encoder or [batch_size, seq_len, emb_dim] for Decoder
        inp_seq_pos: input sequence positions of shape [batch_size, seq_len] 
        '''
        # Prepare non-pad and attention masks
        inp_attn = inp_seq if self.mode == 'Encoder' else inp_seq_pos              # [batch_size, seq_len]
        
        non_pad_mask = (inp_attn.unsqueeze(2) != 0)                                # [batch_size, seq_len, 1]
        attn_mask = (inp_attn.unsqueeze(1) == 0).repeat(1, inp_attn.size(1), 1)    # [batch_size, seq_len, seq_len]

        # Forward
        output = self.embedding(inp_seq) + self.position_encoding(inp_seq_pos)       # [batch_size, seq_len, emb_dim]
        for layer in self.layer_stack:
            output = layer(output, non_pad_mask= non_pad_mask, attn_mask=attn_mask)
        
        output = self.linear(output)
        
        return output


# ### <font color="cyan">3. Length Regulator </font>

# ![](https://i.imgur.com/xUvPwh6.png)

# #### <font color="white"> Duration Predictor </font>
# 

# $PhonemeEmb ⇒ Conv1D ⇒ ReuLU ⇒ LayerNorm ⇒  Dropout ⇒ Conv1D ⇒ ReuLU ⇒ LayerNorm ⇒  Dropout ⇒ Relu ⇒ Linear$

# In[7]:


class DurationPredictor(nn.Module):
    '''
    Duration predictor module. It predicts the duration of each phoneme in the input sequence of encoder phoneme embeddings.
    '''
    def __init__(self, inp_dim, inner_dim, kernel_size, padding_size, dropout=0.1):
        '''
        Initialize the parameters and structure of the duration predictor module.
        inp_dim: input dimension
        inner_dim: inner dimension of the convolutional layers (output dimension of the first convolutional layer)
        kernel_size: kernel size of the convolutional layers
        padding_size: padding size of the convolutional layers
        dropout: dropout probability
        '''
        super(DurationPredictor, self).__init__()

        self.conv1d_1 = nn.Conv1d(inp_dim, inner_dim, kernel_size, padding=padding_size)
        self.layer_norm_1 = nn.LayerNorm(inner_dim)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout)
        self.conv1d_2 = nn.Conv1d(inner_dim, inner_dim, kernel_size, padding=padding_size)
        self.layer_norm_2 = nn.LayerNorm(inner_dim)
        self.linear_layer = nn.Linear(inner_dim, 1)     
        # predicts a scalar mel_duration per phoneme

    def forward(self, encoder_output):
        '''
        Pass given encoder output through the duration predictor module. The input comes from the encoder. 
        encoder_output: encoder output of shape [batch_size, seq_len, emb_dim]
        returns: predicted duration of each phoneme in the input sequence of encoder phoneme embeddings of shape [batch_size, seq_len]
        '''
        x = encoder_output.contiguous().transpose(1, 2)
        x = self.conv1d_1(x)
        x = x.contiguous().transpose(1,2)
        x = self.relu(self.layer_norm_1(x))
        x = self.dropout(x)
        x = x.contiguous().transpose(1, 2)
        x = self.conv1d_2(x)
        x = x.contiguous().transpose(1, 2)
        x = self.relu(self.layer_norm_2(x))
        x = self.dropout(x)
        
        out = self.relu(self.linear_layer(x))

        out = out.squeeze()
        out = out.unsqueeze(0)      # Leading dimension should not be removed in inference.

        return out


# #### <font color="white">Length Regulation Logic </font>
# 

# Given $H=[h_1, h_2, ..., h_n]$ as phone embeddings from the encoder predict the duration of each phoneme $d=[d_1, d_2, ..., d_n]$ and repeat accordingly.

# In[8]:


class LengthRegulator(nn.Module):
    '''
    Length regulator module. It repeats the encoder outputs according to the predicted duration of each phoneme in the input sequence of encoder phoneme embeddings.
    '''
    def __init__(self, inp_dim, inner_dim, kernel_size, padding_size, dropout=0.1):
        '''
        Initialize the parameters and structure of the duration predictor module.
        inp_dim: input dimension
        inner_dim: inner dimension of the convolutional layers (output dimension of the first convolutional layer)
        kernel_size: kernel size of the convolutional layers
        padding_size: padding size of the convolutional layers
        dropout: dropout probability
        '''
        super(LengthRegulator, self).__init__()
        self.duration_predictor = DurationPredictor(inp_dim, inner_dim, kernel_size, padding_size, dropout)

    def forward(self, enc_output):
        '''
        Pass given encoder output through the length regulator module. The input comes from the encoder.
        enc_output: encoder output of shape [batch_size, seq_len, emb_dim]
        '''
        tiny_slowdown = 0.5                                  # to prevent rounding from dropping phonemes 
        duration_predictions = (self.duration_predictor(enc_output) + tiny_slowdown).int()
        # [batch_size, seq_len, enc_dim] to [batch_size, new_seq_len] (scalar mel_duration per phoneme)
        
        # repeat each phoneme in the encoder output according to its predicted duration
        new_seq_lens_per_batch = torch.sum(duration_predictions, -1)
        new_seq_len = torch.max(new_seq_lens_per_batch).item()
        
        batch_size, seq_len, enc_dim = enc_output.size()
        mod_enc_output = torch.zeros((batch_size, new_seq_len, enc_dim))
        for sequence in range(batch_size):
            count = 0
            for phoneme in range(seq_len):
                hidden = enc_output[sequence][phoneme]
                reps = duration_predictions[sequence][phoneme]
                for _ in range(reps):
                    mod_enc_output[sequence][count] = hidden
                    count += 1
        
        # form a new positional encoding for the modified encoder output
        output_pos = torch.LongTensor([i+1 for i in range(mod_enc_output.size(1))]).unsqueeze(0).to(device)
        # [batch_size, seq_len, enc_dim] -> [batch_size, new_seq_len, enc_dim] and [batch_size, new_seq_len]
        return mod_enc_output, output_pos


# ### <font color="cyan">4. Model </font>

# ![](https://i.imgur.com/azh7cwn.png)

# #### Hyperparameters

# In[9]:


# Encoder
VOCAB_SIZE = 300
ENC_EMB_DIM = 256
ENC_NUM_LAYER = 4
ENC_NUM_HEAD = 2
ENC_1D_FILTER_SIZE = 1024
DROPOUT_PROB = 0.1

# Length Regulator
INNER_DIM = 256
DP_KERNEL_SIZE = 3
DROPOUT_PROB = 0.1
DP_PADDING = 1

# Decoder
MAX_SEQ_LEN = 3000
DEC_EMB_DIM = 256
DEC_NUM_LAYER = 4
DEC_NUM_HEAD = 2
DEC_1D_FILTER_SIZE = 1024
MEL_NUM = 80


# #### <font color="white">FastSpeech</font>
# 

# In[10]:


class FastSpeech(nn.Module):
    '''
    FastSpeech 1.0 model. It consists of an encoder, a duration predictor, a length regulator and a decoder.
    '''
    def __init__(self):
        '''
        Initialize the structure of the FastSpeech 1.0 model.
        '''
        super(FastSpeech, self).__init__()

        self.encoder = Dencoder(mode='Encoder', vocab_dim=VOCAB_SIZE, max_seq_len=VOCAB_SIZE, emb_dim=ENC_EMB_DIM,
                                num_layer=ENC_NUM_LAYER, num_head=ENC_NUM_HEAD, h_dim=ENC_EMB_DIM,
                                d_inner=ENC_1D_FILTER_SIZE, mel_num=None, dropout=DROPOUT_PROB)
        
        self.length_regulator = LengthRegulator(inp_dim=ENC_EMB_DIM, inner_dim=INNER_DIM, kernel_size=DP_KERNEL_SIZE, 
                                                padding_size=DP_PADDING, dropout=DROPOUT_PROB)
        
        self.decoder = Dencoder(mode='Decoder', vocab_dim=None, max_seq_len=MAX_SEQ_LEN, emb_dim=DEC_EMB_DIM, 
                                num_layer=DEC_NUM_LAYER, num_head=DEC_NUM_HEAD, h_dim=DEC_EMB_DIM,
                                d_inner=DEC_1D_FILTER_SIZE, mel_num=MEL_NUM,  dropout=DROPOUT_PROB)


    def forward(self, text_seq, src_pos):
        '''
        Pass given input sequence through the FastSpeech 1.0 model. input sequence of shape [batch_size, seq_len] and assigns a unique id to each character in it.
        text_seq: input sequence of shape [batch_size, seq_len]
        src_pos: input sequence positions (indecies) of shape [batch_size, seq_len]
        :returns: predicted mel spectrogram of shape [batch_size, mel_num, new_seq_len]
        '''
        enc_output = self.encoder(text_seq, src_pos)
        
        length_regulator_output, decoder_pos = self.length_regulator(enc_output)
        
        spectogram = self.decoder(length_regulator_output, decoder_pos)

        return spectogram


# ### <font color="cyan">5. Waveglow

# ![](https://i.imgur.com/Xl9HEgm.png)

#  </font>

# ## <font color="yellow"> Inference </font>

# In[11]:


import warnings; warnings.filterwarnings("ignore")
from scipy.io.wavfile import write
try:
    from playsound import playsound
    import waveglow
except:
    pass
import os
import gdown


# In[12]:


class TTS():
    def __init__(self, force_download_wg=False, force_download_fs=False):
        '''
        Initialize the FastSpeech 1.0 model and the WaveGlow model. 
        force_download_wg: whether to force download the WaveGlow weights in case they already seems to exist 
        force_download_fs: whether to force download the FastSpeech 1.0 wieghts in case they already seems to exist
        '''
        # see if there is a WaveGlow folder in the current directory
        # print file and folder names in the current directory
        curr_dir = os.path.dirname(os.path.abspath(__file__))
        if not os.path.exists(curr_dir + '/weights.pth') or force_download_fs:
            print("Weights not found. Downloading FastSpeech 1.0 weights...")
            # if not, download the WaveGlow folder
            f_id = '1G630THkg1CaAZiYAK-rcg7hLNIxu2oO5' 
            gdown.download(f'https://drive.google.com/uc?export=download&confirm=pbef&id={f_id}', curr_dir + '/weights.pth', quiet=False)
            print("Done.")              
        self.symbols = list('_-!\'(),.:;? ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
        self.symbol_to_id = {s: i for i, s in enumerate(self.symbols)}
        self.model = nn.DataParallel(FastSpeech()).to(device)
        # get directory of the current file
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.model.load_state_dict(torch.load(dir_path + "/weights.pth", map_location=device))
        self.model.eval()
        self.WaveGlow = waveglow.load.load_model(download=force_download_wg)
    
    def speak(self, text, play=False, save=False):
        '''
        Pass given text through the FastSpeech 1.0 model and the WaveGlow model to generate speech.
        text: text to be spoken with at most 300 characters
        play: whether to play the generated speech
        save: whether to save the generated speech
        '''
        ascii_text = text.lower()
        sequence = np.array([self.symbol_to_id[s] for s in ascii_text if s in self.symbol_to_id ])
        sequence_inds = np.array([i+1 for w, i in enumerate(sequence)])
        
        sequence, sequence_inds = sequence[np.newaxis, :], sequence_inds[np.newaxis, :]
        
        sequence = torch.from_numpy(sequence).long() if device != torch.device('cuda') else torch.from_numpy(sequence).cuda().long()
        sequence_inds = torch.from_numpy(sequence_inds).long() if device != torch.device('cuda') else torch.from_numpy(sequence_inds).cuda().long()

        with torch.no_grad():
            mel = self.model.module.forward(sequence, sequence_inds)
        mel = mel.contiguous().transpose(1, 2)   
        
        audio = waveglow.inference.get_wav(mel, self.WaveGlow).cpu().numpy()
        if save:
            # get the directory of the current file
            dir_path = os.path.dirname(os.path.realpath(__file__))
            # save the audio file in the current directory
            write(dir_path + "/sample.wav", 22050, audio.astype('int16'))
            
        if play:
            write('sample.wav', 22050, audio.astype('int16'))
            playsound("sample.wav")
            os.remove("sample.wav")
        return audio


# In[13]:


# if running from notebook
if __name__ == '__main__':
    get_ipython().system('jupyter nbconvert --to script FastSpeech.ipynb')




try:
    import os
    from botiverse.models import TTS
    from playsound import playsound
    from botiverse.models import LSTMClassifier
    from botiverse.preprocessors import Vocalize, Wav2Vec,  Frequency
    from botiverse.bots.VoiceBot.utils import voice_input
except:
    pass

class SpeechClassifier():
    '''
    An interface for the speech classifier chatbot which classifies speech into one of a set of classes. Suitable when the
    number of classes is small and the words are easily pronounceable.
    '''
    def __init__(self, words, samplerate, duration, repr='wav2vec', machine='lstm', **kwargs):
        '''
        Initialize the dataset and its transformation for the speech classification process.
        :param words: A list of words which are the classes of the speech classifier.
        :param samplerate: The sample rate of the audio files.
        :param duration: The duration of the audio files in milliseconds.
        '''
        self.words = words
        self.samplerate = samplerate
        self.duration = duration
        self.machine = machine
        if repr == 'wav2vec':
            self.transformer = Wav2Vec(samplerate, duration)
        elif repr == 'mfcc':
            self.transformer = Frequency(type='mfcc', sample_rate=samplerate, duration=duration, **kwargs)
        elif repr == 'spectrogram':
            self.transformer = Frequency(type='spectrogram', sample_rate=samplerate, duration=duration, **kwargs)
        elif type(repr) != str:
            self.transformer = repr
        else:
            raise ValueError(f"Invalid representation {repr}. Expected wav2vec, mfcc or spectrogram.")


    def generate_read_data(self, n=3, regenerate=False, force_download_noise=False, **kwargs):
        '''
        Generate audio data for the words specified during init.
        :param n: The number of audio files to generate for each word.
        '''
        # if there is no dataset folder or if the regenerate flag is set, generate the dataset
        if regenerate or not os.path.exists('dataset'):
            V = Vocalize(self.words)
            Vocalize.corrupt_dataset(self.words, sample_rate=self.samplerate, force_download=force_download_noise)
        X, y = self.transformer.transform_list(self.words, n, **kwargs)
        return X, y

    def fit(self, X, y,  λ=0.001, α=0.01, hidden=128, patience=50, max_epochs=600, **kwargs):
        '''
        Train the speech classifier model.
        :param λ: The learning rate parameter.
        :param α: The regularization parameter.
        :param patience: The number of bad epochs to wait before early stopping.
        :param max_epochs: The maximum number of epochs to train for.
        '''
        if self.machine == 'lstm':
            self.model = LSTMClassifier(X.shape[2], hidden, len(self.words))
            self.model.fit(X, y,  λ, α, max_epochs, patience, **kwargs)
        elif type(self.machine) != str:
            self.model = self.machine
            self.model.fit(X, y, **kwargs)
        else:
            raise ValueError(f"Invalid machine {self.machine}. Expected lstm or a custom model.")
    
    def save(self, path):
        '''
        Save the model to a file.
        :param path: The path to the file
        '''
        self.model.save(path+'.bot')
    
    def load(self, path, **kwargs):
        '''
        Load the model from a file.
        :param path: The path to the file
        '''
        if self.machine == 'lstm':
            self.model = LSTMClassifier(**kwargs)
            self.model.load(path + '.bot')
        else:
            self.model = self.machine
            self.model.load(path + '.bot')
            
        
    
    def predict(self, path, index=False):
        '''
        Predict the class of the audio file at the given path.
        :param path: The path to the audio file to be classified.
        :param index: Whether to return the index of the class or the class itself.
        :return: The class of the audio file at the given path.
        '''
        vec = self.transformer.transform(path, strict_duration=False)
        pred, prob = self.model.predict(vec)
        pred, prob = pred[0], prob[0]
        print("Probability of prediction: ", prob)
        return (self.words[pred], prob) if not index else (pred, prob)



from botiverse.preprocessors.Vocalize.Vocalize import Vocalize
from botiverse.preprocessors.Wav2Vec.Wav2Vec import Wav2Vec
from botiverse.preprocessors.Wav2Vec.Wav2Vec import Wav2Text
from botiverse.preprocessors.BertEmbeddings.BertEmbeddings import BertEmbedder, BertSentenceEmbedder
from botiverse.preprocessors.GloVe.GloVe import GloVe
from botiverse.preprocessors.TF_IDF.TF_IDF import TF_IDF
from botiverse.preprocessors.TF_IDF_GLOVE.TF_IDF_GLOVE import TF_IDF_GLOVE
from botiverse.preprocessors.BoW.BoW import BoW
from botiverse.preprocessors.Frequency.Frequency import Frequency


"""
This Module has the TRIPPY model.
"""

import torch
import torch.nn as nn
from transformers import BertModel

from botiverse.models.BERT.BERT import Bert
from botiverse.models.BERT.config import BERTConfig
from botiverse.models.TRIPPY.config import TRIPPYConfig
from botiverse.models.BERT.utils import LoadPretrainedWeights

class TRIPPY(nn.Module):
  """
  TRIPPY (Task-oriented Reasoning and Inference for Pre-trained models with Pre-trained Ypesystem) model.

  This class implements the TRIPPY model for task-oriented dialogue understanding and slot filling.

  :param n_slots: The number of slots, corresponding to the number of dialogue slots to be filled.
  :type n_slots: int
  :param hid_dim: The hidden dimension size.
  :type hid_dim: int
  :param n_oper: The number of operations.
  :type n_oper: int
  :param dropout: The dropout rate.
  :type dropout: float
  :param from_scratch: Whether to build the BERT model from scratch or load pre-trained weights, defaults to False.
  :type from_scratch: bool
  :param BERT_config: The configuration for the BERT model, defaults to BERTConfig().
  :type BERT_config: BERTConfig
  """

  def __init__(self, n_slots, hid_dim, n_oper, dropout, from_scratch, BERT_config=BERTConfig(), TRIPPY_config=TRIPPYConfig()):
    super(TRIPPY, self).__init__()

    self.hid_dim = hid_dim
    self.n_oper = n_oper
    self.n_slots = n_slots

    if from_scratch == True:
        # Build a BERT model from scratch
        self.bert = Bert(BERT_config)
        LoadPretrainedWeights(self.bert)
    else:
        self.bert = BertModel.from_pretrained('bert-base-uncased')

    aux_dim = 2 * n_slots
    self.oper_layers = nn.ModuleList([nn.Linear(hid_dim + aux_dim, n_oper) for _ in range(n_slots)])
    self.span_layers = nn.ModuleList([nn.Linear(hid_dim, 2) for _ in range(n_slots)])
    self.refer_layers = nn.ModuleList([nn.Linear(hid_dim + aux_dim, n_slots + 1) for _ in range(n_slots)])
    self.dropout = nn.Dropout(dropout)

    # auxiliary features layers
    self.inform_aux_layer = nn.Linear(n_slots, n_slots)
    self.ds_aux_layer = nn.Linear(n_slots, n_slots)

  def forward(self, ids, mask, token_type_ids, inform_aux_features, ds_aux_features):
    """
    Forward pass of the TRIPPY model.

    :param ids: The input token IDs.
    :type ids: torch.Tensor, shape [batch size, seq len]
    :param mask: The attention mask indicating which tokens are valid.
    :type mask: torch.Tensor, shape [batch size, seq len]
    :param token_type_ids: The token type IDs.
    :type token_type_ids: torch.Tensor, shape [batch size, seq len]
    :param inform_aux_features: The auxiliary features for informing slots.
    :type inform_aux_features: torch.Tensor, shape [batch size, n_slots]
    :param ds_aux_features: The auxiliary features for dialogue state tracking.
    :type ds_aux_features: torch.Tensor, shape [batch size, n_slots]
    :return: Tuple containing the logits for slot start positions, slot end positions, slot operations, and slot references.
    :rtype: tuple(torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor)
    """
    
    # ids = [batch size, seq len]
    # mask = [batch size, seq len]
    # token_type_ids = [batch size, seq len]

    # sequence_output = [batch size, seq len, hid_dim]
    # pooled_output = [batch size, hid_dim]
    sequence_output, pooled_output = self.bert(ids,
                                               attention_mask=mask,
                                               token_type_ids=token_type_ids,
                                               return_dict=False
                                               )

    sequence_output = self.dropout(sequence_output)
    pooled_output = self.dropout(pooled_output)

    # concatenate the auxiliary features
    # pooled_output = [batch size, hid_dim + 2 * n_slots]
    # print(pooled_output.shape)
    # print(inform_aux_features.shape)
    # print(ds_aux_features.shape)
    pooled_output = torch.cat((pooled_output, self.inform_aux_layer(inform_aux_features), self.ds_aux_layer(ds_aux_features)), 1)
    # print(pooled_output.shape)
    # print("\n\n\n\n")

    slots_start_logits = []
    slots_end_logits = []
    slots_oper_logits = []
    slots_refer_logits = []
    for slot in range(self.n_slots):

      # oper_logits = [batch size, n_oper]
      oper_logits = self.oper_layers[slot](pooled_output)

      # span_logits = [batch size, seq len, 2]
      span_logits = self.span_layers[slot](sequence_output)

      # start_logits = [batch size, seq len, 1]
      # end_logits = [batch size, seq len, 1]
      start_logits, end_logits = span_logits.split(1, dim=-1)

      # start_logits = [batch size, seq len]
      # end_logits = [batch size, seq len]
      start_logits = start_logits.squeeze(-1)
      end_logits = end_logits.squeeze(-1)

      # refer_logits = [batch size, n_slots + 1]
      refer_logits = self.refer_layers[slot](pooled_output)

      slots_start_logits.append(start_logits)
      slots_end_logits.append(end_logits)
      slots_oper_logits.append(oper_logits)
      slots_refer_logits.append(refer_logits)

    return slots_start_logits, slots_end_logits, slots_oper_logits, slots_refer_logits


try:
    import torchaudio
    import os
    from tqdm import tqdm
    import torch
    import librosa
    import numpy as np
    import matplotlib.pyplot as plt
    from torch.utils.data import Dataset
    from audiomentations import Compose, AddGaussianNoise, TimeStretch, PitchShift, Shift
except:
    pass
class Frequency():
    '''
    An interface for transforming audio files into frequency domain representations.
    '''
    def __init__(self, sample_rate=16000, duration=1, augment=None, type='spec', 
                 nmels=70, n_fft=720, hop_length=360, is_log=True, **kwargs):

        self.sample_rate = sample_rate
        self.duration = duration
        self.emb_dim = 768 * int(49 * self.duration)         # fact regarding wav2vec2-base-960h
        if augment is None:
            self.augment = Compose([
            TimeStretch(min_rate=0.8, max_rate=1.4, p=0.7),
            PitchShift(min_semitones=-4, max_semitones=1, p=0.8),
            Shift(min_fraction=-0.5, max_fraction=0.5, p=0.5),
            AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.015, p=0.5),
        ])
        else:
            self.augment = augment
        
        if type == 'spectrogram':
            self.transform_func = torchaudio.transforms.MelSpectrogram(sample_rate=self.sample_rate, n_fft=n_fft, hop_length=hop_length, n_mels=nmels, **kwargs)
        elif type == 'mfcc':
            self.transform_func = torchaudio.transforms.MFCC(sample_rate=self.sample_rate, n_mfcc=nmels, melkwargs={'n_fft':n_fft, 'hop_length':hop_length}, **kwargs)
        else:   raise ValueError(f"Invalid type of transformation {type}. Expected spectrogram or mfcc")
        
        self.is_log = is_log

    def transform_list(self, words, n=4):
        '''
        Given a folder dataset with folders each containing audio files, this returns a table of spectra in the form of a numpy array X and a table of classes in the form of a numpy array y.
        :param: words: A list of words which are the classes of the speech classifier.
        :param: n: The number of times to augment each audio file.
        '''        
        # may be needed in the future (draft)
        sounds_per_word = len(os.listdir(f"dataset/{words[0]}"))
        self.N = len(words) * sounds_per_word
        
        # main function
        x_data, y_data = [], []
        for word in tqdm(words):
            for i, file in enumerate(os.listdir(f"dataset/{word}")):
                if not file.endswith(".wav"): continue
                waveform, sr = torchaudio.load(f"dataset/{word}/{file}")
                waveform = torchaudio.transforms.Resample(sr, self.sample_rate)(waveform)

                if waveform.shape[0] == 2:
                    waveform = torch.mean(waveform, dim=0, keepdim=True)
                
                length = waveform.shape[1]
                sample_dur = int(self.duration * self.sample_rate)
                if length < sample_dur:
                    waveform = torch.cat((waveform, torch.zeros(1, sample_dur - length)), dim=1)
                elif length > sample_dur:
                    waveform = waveform[:, :sample_dur]     
                
                waveform = waveform.numpy()
                if n > 0:
                    for _ in range(n):
                        waveform = self.augment(waveform, sample_rate=self.sample_rate)
                        spectrum = self.transform_func(torch.from_numpy(waveform))
                        spectrum = librosa.power_to_db(spectrum[0]) if self.is_log else spectrum[0]
                        # transpose to get (time, freq) instead of (freq, time)
                        spectrum = spectrum.T
                        x_data.append(spectrum)
                        y_data.append(words.index(word))                
                else:
                    spectrum = self.transform_func(torch.from_numpy(waveform))
                    spectrum = librosa.power_to_db(spectrum[0]) if self.is_log else spectrum[0]
                    # transpose to get (time, freq) instead of (freq, time)
                    spectrum = spectrum.T
                    x_data.append(spectrum)
                    y_data.append(words.index(word))
                    
        x_data, y_data = np.array(x_data), np.array(y_data)
        
        return x_data, y_data


    
    def transform(self, path, strict_duration=False):
        '''
        Convert the audio file given in path into a frequency domain representation.
        '''
        waveform, sr = torchaudio.load(path)
        waveform = torchaudio.transforms.Resample(sr, self.sample_rate)(waveform)
        if waveform.shape[0] == 2:  waveform = torch.mean(waveform, dim=0, keepdim=True)
        length = waveform.shape[1]
        if strict_duration:
            sample_dur = int(self.duration * self.sample_rate)
            if length < sample_dur:
                waveform = torch.cat((waveform, torch.zeros(1, sample_dur - length)), dim=1)
            elif length > sample_dur:
                waveform = waveform[:, :sample_dur]
            
        spectrum = self.transform_func(waveform)
        spectrum = librosa.power_to_db(spectrum[0].numpy()) if self.is_log else spectrum[0].numpy()
        spectrum = spectrum.T
        spectrum = spectrum[np.newaxis, ...]
        return spectrum
        



'''
def plot_freq_domains(words, spectograms, num_pos=SOUNDS_PER_WORD):
    fig, axs = plt.subplots(len(words), num_pos, figsize=(15, 4*len(words)))
    for i, word in enumerate(words):
        for j, (prop, mel) in enumerate(spectograms[word].items()):
            axs[i, j].imshow(mel, origin="lower", aspect="auto")
            axs[i, j].set_title(f"{word.upper()}-{prop}")        
            if j==0:   
                axs[i, j].set_ylabel("freq_bin")
                axs[i, j].set_xlabel("time_window")
        fig.subplots_adjust(hspace=0.5)
    plt.show(block=False)
    

def random_waveform(words, mel_transform, mfcc_transform):
    # select random word from words
    word = np.random.choice(words)
    # select random audio file from word folder
    file = np.random.choice(os.listdir('dataset/' + word))
    # load audio file
    waveform, sample_rate = torchaudio.load('dataset/' + word + '/' + file)
    mel_specgram = mel_transform(waveform)
    mfcc_specgram = mfcc_transform(waveform)
    # plot them side by side
    plt.rcParams['figure.dpi'] = 125
    plt.style.use('dark_background')
    fig, axs = plt.subplots(1, 2, figsize=(15, 4))
    axs[0].imshow(librosa.power_to_db(mel_specgram[0].numpy()), origin="lower", aspect="auto")
    axs[0].set_title("Mel Spectrogram")
    axs[0].set_ylabel("freq_bin")
    axs[0].set_xlabel("time_window")
    axs[1].imshow(librosa.power_to_db(mfcc_specgram[0].numpy()), origin="lower", aspect="auto")
    axs[1].set_title("MFCC")
    axs[1].set_ylabel("freq_bin")
    axs[1].set_xlabel("time_window")
    fig.suptitle(f"\'{word.upper()}\'")
    plt.show(block=False)
    print("Shape of waveform: {}".format(waveform.size()))
    print("Sample rate of waveform: {}".format(sample_rate))



class Spectograms(Dataset):
    def __init__(self, x_data, y_data, is_train=True, transform=None, normalize=False):
        self.is_train = is_train
        self.transform = transform
        self.num_classes = len(np.unique(y_data))
        
        # split the data into train and test
        self.x_train, self.x_val, self.y_train, self.y_val = \
            train_test_split(x_data, y_data, test_size=TEST_SIZE, random_state=1, stratify=y_data)
            
        if normalize:
            self.μ, self.σ = np.mean(self.x_train), np.std(self.x_train)
            self.x_train = (self.x_train - self.μ) / self.σ
            self.x_val = (self.x_val - self.μ) / self.σ
            
    def __len__(self):
        return len(self.x_train) if self.is_train else len(self.x_val)

    def __getitem__(self, index):
        if self.is_train:
            a, y = self.x_train[index], self.y_train[index]
            a, y = self.x_train[index], self.y_train[index]
            # Get the list of images where self.y_train == y and exclude the anchor "a"
            p_ind = np.where(self.y_train == y)[0]
            # exclude the anchor which has ind as index
            p_ind = np.delete(p_ind, np.argwhere(p_ind == index))
            # pop index of the anchor from the list
            p = self.x_train[p_ind[random.randrange(0, p_ind.shape[0])]]

            n_ind = np.where(self.y_train != y)[0]
            # randomly select np.unique(self.y_train).shape[0] number of images
            n_ind = random.sample(list(n_ind), int(NUM_NEGATIVES*self.num_classes))
            list_n = self.x_train[n_ind]
            # find a random negative image
            #n = random.choice(self.x_train[self.y_train != y])

            if self.transform:
                # convert each image to PIL in list_n
                list_n = [self.transform(n) for n in list_n]
                list_n = torch.stack(list_n)
                a, p = a[...,np.newaxis], p[...,np.newaxis]
                a, p = self.transform(a), self.transform(p)

            return a, p, list_n, y

        else:
            a, y = self.x_val[index], self.y_val[index]
            if self.transform:
                a = a[...,np.newaxis]
                a = self.transform(a)
            return a, y
        
 
def show_next_triplet(train_loader, words):
    a, p, list_n, y = next(iter(train_loader))
    a0, p0, list_n0, y0 = a[0].squeeze(), p[0].squeeze(), list_n[0].squeeze(), y[0].squeeze()
    fig, ax = plt.subplots(1, 2+len(list_n0), figsize=(0.5*(2+len(list_n0))*5, 2))
    ax[0].imshow(a0, origin="lower", aspect="auto")
    ax[0].set_title(f'Anchor: {words[y[0]]}')
    ax[1].imshow(p0, origin="lower", aspect="auto")
    ax[1].set_title('Positive')
    for i in range(len(list_n0)):
        ax[i+2].imshow(list_n0[i].squeeze(), origin="lower", aspect="auto")
        ax[i+2].set_title(f'Negative')
    plt.show()
    
'''


from botiverse.bots.BasicBot.BasicBot import BasicBot
from botiverse.bots.WhizBot.WhizBot import WhizBot
from botiverse.bots.ConverseBot.ConverseBot import ConverseBot
from botiverse.bots.BasicTaskBot.BasicTaskBot import BasicTaskBot
from botiverse.bots.TaskBot.TaskBot import TaskBot
from botiverse.bots.VoiceBot.SpeechClassifier import SpeechClassifier
from botiverse.bots.VoiceBot.VoiceBot import VoiceBot



"""
This Module contains the data processing functions for TRIPPY.
"""

import json
import torch
import numpy as np
import re
from tqdm import tqdm

from botiverse.models.TRIPPY.utils import RawDataInstance, DataInstance, normalize, is_included, included_with_label_maps, match_with_label_maps, mask_utterance

def fix_slot_list(slot_list, domains):
  """
  Fix slot list by filtering slots based on domains.

  :param slot_list: List of slot names.
  :type slot_list: list[str]

  :param domains: The list of domains to filter the slots.
  :type domains: list[str]

  :return: The new fixed and sorted slot list.
  :rtype: list[str]
  """

  # delete slots not in the domains
  del_slot = []
  for slot in slot_list:
    found = False
    for domain in domains:
      if domain in slot:
        found = True
    if found == False:
      del_slot.append(slot)
  for slot in del_slot:
    del slot_list[slot_list.index(slot)]

  return sorted(slot_list)


def read_raw_data(data_path, slot_list, max_len, domains, multiwoz):
  """
  Read raw data from the JSON file and preprocess it.

  :param data_path: The path to the JSON data file.
  :type data_path: str

  :param slot_list: The list of slots.
  :type slot_list: list[str]

  :param max_len: The maximum length of the input sequence.
  :type max_len: int

  :param domains: The list of domains.
  :type domains: list[str]

  :return: The list of raw data instances.
  :rtype: list[RawDataInstance]
  """

  # read data
  file = open(data_path)
  parsed_data = json.load(file)

  raw_data = []
  # loop over dialogues
  for dial_info in parsed_data:
    dial_idx = dial_info['dialogue_idx']

    history = []
    # loop over dialogue turns
    for turn in dial_info['dialogue']:

      # turn id
      turn_idx = turn['turn_idx']

      # turn utterances
      user_utter = turn['user_utterance']
      sys_utter = turn['system_utterance']

      # normalize utterances
      user_utter = ' '.join(normalize(user_utter, multiwoz))
      sys_utter = ' '.join(normalize(sys_utter, multiwoz))

      # get the changed slots in this turn
      turn_slots = turn['turn_slots']

      # Get system actions which will be used as the inform memory
      inform_mem = turn['system_act']

      # mask the system utterance by removing labels appeared in system acts
      sys_utter = ' '.join(mask_utterance(sys_utter, inform_mem, multiwoz, '[UNK]'))

      # append current instance
      raw_data.append(RawDataInstance(dial_idx,
                                      turn_idx,
                                      user_utter,
                                      sys_utter,
                                      history,
                                      turn_slots,
                                      inform_mem))

      # update history & last state for next turn
      history = [user_utter, sys_utter] + history

  return raw_data


def create_slot_span(input, target_value, tok_input_offsets, padding_len, label_maps):
  """
  Create a slot span given the input, target value, and token input offsets, 
  by matching the target value as tokens with the input sequence. 

  :param input: The input string.
  :type input: str

  :param target_value: The target value.
  :type target_value: str

  :param tok_input_offsets: The token input offsets.
  :type tok_input_offsets: list[tuple[int, int]]

  :param padding_len: The padding length.
  :type padding_len: int

  :param label_maps: The label maps.
  :type label_maps: dict

  :return: The slot span, span start index, and span end index.
  :rtype: tuple[list[int], int, int]
  """

  # get all possible variants of the slot value
  label_variants = [target_value]
  if target_value in label_maps:
    label_variants = label_variants + label_maps[target_value]

  # match the target value as tokens
  start, end = -1, -1
  found = False
  input_list = input.split()
  first_idx = input_list.index('[SEP]')
  max_idx = first_idx
  for label in label_variants:
    label_list = [item for item in map(str.strip, re.split("(\W+)", label)) if len(item) > 0]
    if found == True:
      break
    for idx in (j for j, e in enumerate(input_list) if(e == label_list[0] and j < max_idx)):
      if input_list[idx:idx + len(label_list)] == label_list:
        start, end = idx, idx + len(label_list) - 1
        found = True

  # mark the selected part as characters in the input
  input = " ".join(input_list)
  ch_start, ch_end = -1, -1
  acc_len = 0
  for idx, tok in enumerate(input_list):
    if start == idx:
      ch_start = acc_len + idx
    acc_len += len(tok)
    if end == idx:
      ch_end = acc_len + idx - 1

  # mark the target span in the input string
  char_target = [0] * len(input)
  if ch_start != -1 and ch_end != -1:
    for j in range(ch_start, ch_end + 1):
      if input[j] != " ":
        char_target[j] = 1

  # mark the target span after tokenization
  span = [0] * len(tok_input_offsets)
  for j, (offset1, offset2) in enumerate(tok_input_offsets):
    if sum(char_target[offset1:offset2]) > 0:
      span[j] = 1

  # update the target as tok_input_offsets doesn not include
  # [CLS] & [SEP] in the start & end of input string
  span = [0] + span + [0]

  # get the start & end index of the span if any
  # otherwise 0
  span_start = 0
  span_end = 0
  non_zero = np.nonzero(span)[0]
  if len(non_zero) > 0:
    span_start = non_zero[0]
    span_end = non_zero[-1]

  # pad the target span
  span = span + [0] * padding_len

  return span, span_start, span_end


def create_inputs(history, user_utter, sys_utter, tokenizer, max_len):
  """
  Create inputs for BERT using the history, user utterance, system utterance,
  by creating and tokenizing the input seqence and creating the masks.

  :param history: The history of utterances.
  :type history: list[str]

  :param user_utter: The user's utterance.
  :type user_utter: str

  :param sys_utter: The system's utterance.
  :type sys_utter: str

  :param tokenizer: The tokenizer to tokenize the input.
  :type tokenizer: transformers.PreTrainedTokenizer

  :param max_len: The maximum length of the input.
  :type max_len: int

  :return: The input string, token IDs, attention mask, token type IDs,
            token input offsets, tokenized input tokens, and padding length.
  :rtype: tuple[str, list[int], list[int], list[int], list[tuple[int, int]],
                list[str], int]
  """


  # create input string
  history = " ".join(history)
  current_utter = user_utter + ' [SEP] ' + sys_utter + ' [SEP] '
  input = current_utter + history
  input = " ".join(input.split())

  # tokenize and truncate input
  tok_input = tokenizer.encode(input)
  tok_input_tokens = tok_input.tokens[:max_len]
  tok_input_ids = tok_input.ids[:max_len]
  tok_input_offsets = tok_input.offsets[1:max_len-1]
  if tok_input_tokens[-1] != '[SEP]':
    tok_input_tokens[-1] = '[SEP]'
    tok_input_ids[-1] = 102

  # create mask & input type id
  mask = [1] * len(tok_input_ids)
  token_type_ids = []
  cnt = 0
  for i, token in enumerate(tok_input_tokens):
    token_type_ids.append(1 if cnt >= 2 else 0)
    cnt += 1 if token == '[SEP]' else 0

  # pad the inputs
  padding_len = max_len - len(tok_input_ids)
  ids = tok_input_ids + [0] * padding_len
  mask = mask + [0] * padding_len
  token_type_ids = token_type_ids + [0] * padding_len

  return input, ids, mask, token_type_ids, tok_input_offsets, tok_input_tokens, padding_len


def is_informed(value, target, label_maps, multiwoz):
  """
  Check if a value is informed by the system given a target value and label maps.

  :param value: The value to check.
  :type value: str

  :param target: The target value.
  :type target: str

  :param label_maps: The label maps.
  :type label_maps: dict

  :return: A tuple indicating if the value is informed and the informed value.
  :rtype: tuple[bool, str]
  """

  informed = False
  informed_value = 'none'

  target = ' '.join(normalize(target, multiwoz))

  if value == target or is_included(value, target) or is_included(target, value):
    informed = True
  if value in label_maps:
    informed = included_with_label_maps(target, value, label_maps)
  elif target in label_maps:
    informed = included_with_label_maps(value, target, label_maps)
  if informed: informed_value = value

  return informed, informed_value


def get_refered_slot(target_value, slot, last_state, non_referable_slots, non_referable_pairs, label_maps={}):
    """
    Get the referred slot if the user refers to another slot in the dialogue state given a target value, slot, last state, 
    non-referable slots, non-referable pairs, and label maps.

    :param target_value: The target value.
    :type target_value: str

    :param slot: The slot to check.
    :type slot: str

    :param last_state: The last state.
    :type last_state: dict

    :param non_referable_slots: The list of non-referable slots.
    :type non_referable_slots: list[str]

    :param non_referable_pairs: The list of non-referable slot pairs.
    :type non_referable_pairs: list[tuple[str, str]]

    :param label_maps: The label maps.
    :type label_maps: dict, optional

    :return: The referred slot.
    :rtype: str
    """

    referred_slot = 'none'

    if slot in non_referable_slots:
        return referred_slot

    if slot in last_state and last_state[slot] == target_value:
      return referred_slot

    for s in last_state:

        if s in non_referable_slots:
            continue

        if ((slot, s) in non_referable_pairs) or ((s, slot) in non_referable_pairs):
          continue

        if slot == s:
          continue

        if match_with_label_maps(last_state[s], target_value, label_maps):
            referred_slot = s
            break

    return referred_slot


def create_labels(target_value, slot, last_state, input, tok_input_offsets, inform_mem, label_maps, padding_len, max_len, non_referable_slots, non_referable_pairs, multiwoz):
  """
  Create the target operation and the span labels for a slot.

  :param target_value: The target value.
  :type target_value: str

  :param slot: The slot.
  :type slot: str

  :param last_state: The last state.
  :type last_state: dict

  :param input: The input string.
  :type input: str

  :param tok_input_offsets: The token input offsets.
  :type tok_input_offsets: list[tuple[int, int]]

  :param inform_mem: The inform memory.
  :type inform_mem: dict

  :param label_maps: The label maps.
  :type label_maps: dict

  :param padding_len: The padding length.
  :type padding_len: int

  :param max_len: The maximum length of the input.
  :type max_len: int

  :param non_referable_slots: The list of non-referable slots (slots that can not use refering).
  :type non_referable_slots: list[str]

  :param non_referable_pairs: The list of non-referable slot pairs (slots pairs that can not refer to each other).
  :type non_referable_pairs: list[tuple[str, str]]

  :return: The operation, span, span start index, span end index, referred slot, and informed value.
  :rtype: tuple[str, list[int], int, int, str, str]
  """

  oper = 'carryover'
  span = [0] * max_len
  span_start = 0
  span_end = 0
  refered_slot = 'none'
  informed_value = 'none'

  # assert target_value != 'none', 'target value can not be none'

  if target_value in ['[NULL]', 'none']:
    oper = 'carryover'
  elif target_value in ['dontcare', 'yes', 'no']:
    oper = target_value
  else:
    span, span_start, span_end = create_slot_span(input,
                                                  target_value,
                                                  tok_input_offsets,
                                                  padding_len,
                                                  label_maps)

    informed = False
    if slot in inform_mem:
      assert len(inform_mem[slot]) == 1, 'greater than 1'
      informed, informed_value = is_informed(inform_mem[slot][0], target_value, label_maps, multiwoz)

    refered_slot = get_refered_slot(target_value, slot, last_state, non_referable_slots, non_referable_pairs, label_maps)

    if sum(span) != 0:
      oper = 'update'
    elif informed == True:
      oper = 'inform'
    elif refered_slot != 'none':
      oper = 'refer'
    else:
      oper = 'unpointable'

  return oper, span, span_start, span_end, refered_slot, informed_value


def create_data(raw_data, slot_list, label_maps, tokenizer, max_len, non_referable_slots, non_referable_pairs, multiwoz):
  """
  Create the data instances for training or evaluation.

  :param raw_data: The list of raw data instances.
  :type raw_data: list[RawDataInstance]

  :param slot_list: The list of slots.
  :type slot_list: list[str]

  :param label_maps: The label maps.
  :type label_maps: dict

  :param tokenizer: The tokenizer to tokenize the input.
  :type tokenizer: transformers.PreTrainedTokenizer

  :param max_len: The maximum length of the input.
  :type max_len: int

  :param non_referable_slots: The list of non-referable slots.
  :type non_referable_slots: list[str]

  :param non_referable_pairs: The list of non-referable slot pairs.
  :type non_referable_pairs: list[tuple[str, str]]

  :return: The list of data instances.
  :rtype: list[DataInstance]
  """

  data = []

  last_state = {}
  cur_state = {}
  prev_dial_idx = -1
  # loop over raw data
  for turn in tqdm(raw_data):

    # if new dialogue reset the state
    if turn.dial_idx != prev_dial_idx or turn.turn_idx == 0:
      cur_state = {}
      last_state = {}

    # update previous dialogue index
    prev_dial_idx = turn.dial_idx

    # create model inputs
    input, ids, mask, token_type_ids, tok_input_offsets, input_tokens, padding_len = create_inputs(turn.history,
                                                                                                   turn.user_utter,
                                                                                                   turn.sys_utter,
                                                                                                   tokenizer,
                                                                                                   max_len)


    target_values = []
    opers = []
    spans = []
    spans_start = []
    spans_end = []
    refer = ['none'] * len(slot_list)
    inform_aux_features = [0] * len(slot_list)
    ds_aux_features = [0] * len(slot_list)

    # for each slot determine its values
    for slot_idx, slot in enumerate(slot_list):

      # get the slot target value
      target_value = '[NULL]'
      if slot in turn.turn_slots:
        target_value = turn.turn_slots[slot]
      elif slot in cur_state:
        target_value = cur_state[slot]


      # get slot labels
      (oper,
       span,
       span_start,
       span_end,
       refered_slot,
       informed_value) = create_labels(target_value,
                                      slot,
                                      last_state,
                                      input,
                                      tok_input_offsets,
                                      turn.inform_mem,
                                      label_maps,
                                      padding_len,
                                      max_len,
                                      non_referable_slots,
                                      non_referable_pairs,
                                      multiwoz)

      if slot in cur_state and target_value == cur_state[slot] and oper in ['dontcare', 'yes', 'no', 'refer']:
        oper = 'carryover'


      # create auxiliary features
      # mark each informed slot as 1
      if slot in turn.inform_mem:
        inform_aux_features[slot_idx] = 1
      # mark each filled slot as 1
      if slot in cur_state:
        ds_aux_features[slot_idx] = 1

      # update the state
      if oper != 'carryover':
        cur_state[slot] = target_value
        if oper == 'unpointable':
          oper = 'carryover'

#       if turn.dial_idx == 'MUL2491.json' and turn.turn_idx == 8 and slot == 'restaurant-name':
#         print(oper)
#         print(span)
#         print(refered_slot)
#         print(informed_value)
#         print(last_state)
#         print(cur_state)

      target_values.append(target_value)
      opers.append(oper)
      spans.append(span)
      spans_start.append(span_start)
      spans_end.append(span_end)
      refer[slot_idx] = refered_slot

    data.append(DataInstance(ids,
                             mask,
                             token_type_ids,
                             spans,
                             spans_start,
                             spans_end,
                             padding_len,
                             input_tokens,
                             input,
                             opers,
                             target_values,
                             last_state.copy(),
                             cur_state.copy(),
                             refer,
                             inform_aux_features,
                             ds_aux_features))

    # update last state
    last_state = cur_state.copy()


  return data


def prepare_data(data_path, slot_list, label_maps, tokenizer, max_len, domains, non_referable_slots, non_referable_pairs, multiwoz):
  """
  Prepare the data for training or evaluation, this usually the function you want to call to preprocess the data for
  TripPy model, it encapsulates the whole process of preprcessing the data by calling the other functions in this
  module.

  :param data_path: The path to the JSON data file.
  :type data_path: str

  :param slot_list: The list of slots.
  :type slot_list: list[str]

  :param label_maps: The label maps.
  :type label_maps: dict

  :param tokenizer: The tokenizer to tokenize the input.
  :type tokenizer: transformers.PreTrainedTokenizer

  :param max_len: The maximum length of the input.
  :type max_len: int

  :param domains: The list of domains.
  :type domains: list[str]

  :param non_referable_slots: The list of non-referable slots.
  :type non_referable_slots: list[str]

  :param non_referable_pairs: The list of non-referable slot pairs.
  :type non_referable_pairs: list[tuple[str, str]]

  :return: The raw data and prepared data.
  :rtype: tuple[list[RawDataInstance], list[DataInstance]]
  """


  # create raw data
  raw_data = read_raw_data(data_path, slot_list, max_len, domains, multiwoz)

  # create data
  data = create_data(raw_data, slot_list, label_maps, tokenizer, max_len, non_referable_slots, non_referable_pairs, multiwoz)

  return raw_data, data


class Dataset(torch.utils.data.Dataset):
  """
  PyTorch Dataset for the TRIPPY model.
  
  :param data: The list of data instances.
  :type data: list[DataInstance]

  :param n_slots: The number of slots.
  :type n_slots: int

  :param oper2id: The mapping of operations to IDs.
  :type oper2id: dict[str, int]

  :param slot_list: The list of slots.
  :type slot_list: list[str]
  """

  def __init__(self, data, n_slots, oper2id, slot_list):

    # for k in inputs:
    #   inputs[k] = inputs[k][:32]

    self.ids = [turn.ids for turn in data]
    self.mask = [turn.mask for turn in data]
    self.token_type_ids = [turn.token_type_ids for turn in data]
    self.spans_start = [turn.spans_start for turn in data]
    self.spans_end = [turn.spans_end for turn in data]
    self.padding_len = [turn.padding_len for turn in data]
    self.input_tokens = [' '.join(turn.input_tokens) for turn in data]
    self.target_values = ['[VALUESEP]'.join(turn.target_values) for turn in data]
    self.opers = [[oper2id[oper] for oper in turn.opers] for turn in data]
    # get the index of the refered slot, in case the slot is not present in the slot_list then that means "none"
    # index of "none" is n_slots
    self.refer = [[(slot_list.index(r) if r in slot_list else n_slots) for r in turn.refer] for turn in data]
    self.inform_aux_features = [turn.inform_aux_features for turn in data]
    self.ds_aux_features = [turn.ds_aux_features for turn in data]


  def __len__(self):
    """
    Get the length of the dataset.

    :return: The length of the dataset.
    :rtype: int
    """
    return len(self.ids)

  def __getitem__(self, idx):
    """
    Get an item from the dataset at the given index.

    :param idx: The index of the item.
    :type idx: int

    :return: The item at the given index.
    :rtype: dict[str, torch.Tensor or str]
    """
    return {
        'ids': torch.tensor(self.ids[idx], dtype=torch.long),
        'mask': torch.tensor(self.mask[idx], dtype=torch.long),
        'token_type_ids': torch.tensor(self.token_type_ids[idx], dtype=torch.long),
        'spans_start': torch.tensor(self.spans_start[idx], dtype=torch.long),
        'spans_end': torch.tensor(self.spans_end[idx], dtype=torch.long),
        'padding_len': torch.tensor(self.padding_len[idx], dtype=torch.long),
        'input_tokens': self.input_tokens[idx],
        'target_values': self.target_values[idx],
        'opers': torch.tensor(self.opers[idx], dtype=torch.long),
        'refer': torch.tensor(self.refer[idx], dtype=torch.long),
        'inform_aux_features': torch.tensor(self.inform_aux_features[idx], dtype=torch.float),
        'ds_aux_features': torch.tensor(self.ds_aux_features[idx], dtype=torch.float)
    }


"""
This file is used to build the package and upload it to PyPI
"""
# pylint: skip-file  (PyLint is not happy with importing the already built-in open)
from os import path

# To use a consistent encoding
from codecs import open

# Always prefer setuptools over distutils
from setuptools import setup


# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

# requirements.txt
def read_requirements(name):
    """Read requirements from requirements.txt."""
    with open(path.join(HERE, name), encoding='utf-8') as f:
        requirements = f.read().splitlines()
    return requirements



# This call to setup() does all the work
setup(
    name="botiverse",
    version="0.4.11",
    description='''botiverse is a chatbot library that offers a high-level API to access a diverse set of chatbot models''',
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://botiverse.readthedocs.io/",
    author="Essam W., Mohamed Saad, Yousef Atef, Karim Taha",
    author_email="essamwisam@outlook.com",
    license="GPLv3",
    classifiers=[                               # https://pypi.org/classifiers/
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    package_dir={
        "botiverse": "botiverse",
        "botiverse.bots": "botiverse/bots",
        "botiverse.models": "botiverse/models",
        "botiverse.preprocessors": "botiverse/preprocessors",
    },
    packages=["botiverse", 
              "botiverse.bots", 
              "botiverse.bots.BasicBot", 
              "botiverse.bots.ConverseBot", 
              "botiverse.bots.VoiceBot",
              "botiverse.bots.WhizBot", 
              "botiverse.bots.BasicTaskBot",
              "botiverse.bots.TaskBot", 
              "botiverse.models",
              "botiverse.preprocessors", 
              "botiverse.gui",
              "botiverse.models.BERT",
              "botiverse.models.FastSpeech1",
              "botiverse.models.GRUClassifier",
              "botiverse.models.LinearClassifier",
              "botiverse.models.LSTM",
              "botiverse.models.NN",
              "botiverse.models.SVM",
              "botiverse.models.T5Model",
              "botiverse.models.TRIPPY",
              "botiverse.preprocessors.BertEmbeddings",
              "botiverse.preprocessors.BoW",
              "botiverse.preprocessors.Frequency",
              "botiverse.preprocessors.GloVe",
              "botiverse.preprocessors.Special",
              "botiverse.preprocessors.Special.ConverseBot_Preprocessor",
              "botiverse.preprocessors.Special.WhizBot_BERT_Preprocessor",
              "botiverse.preprocessors.Special.WhizBot_GRU_Preprocessor",
              "botiverse.preprocessors.GloVe",    
              "botiverse.preprocessors.TF_IDF",
              "botiverse.preprocessors.TF_IDF_GLOVE",
              "botiverse.preprocessors.Vocalize",
              "botiverse.preprocessors.Wav2Vec",
              "botiverse.gui.static",
              "botiverse.gui.static.icons",
              "botiverse.gui.templates",
              ],
        include_package_data = True,
        package_data={
            'botiverse.gui':['*.zip', '*.png', '*pdf', '*jpeg','*ipynb', '*html', '*css', '*pkl', '*js'],
            'botiverse.gui.static':['*.zip', '*.png', '*pdf', '*jpeg','*ipynb', '*html', '*css', '*pkl', '*js'],
            'botiverse.gui.static.icons':['*.zip', '*.png', '*pdf', '*jpeg','*ipynb', '*html', '*css', '*pkl', '*js'],
            'botiverse.gui.templates':['*.zip', '*.png', '*pdf', '*jpeg','*ipynb', '*html', '*css', '*pkl', '*js'],
            },
    install_requires=read_requirements('requirements.txt'),
    extras_require={
        "voice": read_requirements("requirements_voice.txt"),
    }
)


# Steps to upload to PyPI
# 0 - Increment the version number in setup.py
# 1 - Remove the dist folder
# 2- python3 setup.py sdist bdist_wheel  
# 3 - twine upload dist/*

# To upload to test PyPI: twine upload --repository-url https://test.pypi.org/legacy/ dist/*



"""
This module ontains utility code used by the deep Taskbot module
such as Natural Language Understanding (NLU), Dialogue State Tracker (DST) ... etc.
"""

import random

class PriorityDP:
    """
    Dialogue Policy Optimizer that selects the next action based on priority.

    This policy selects the action with the highest priority that does not conflict with
    the slots already filled in the dialogue state.

    """
    def get_action(self, state, templates_slots):
        """
        Get the next action based on the given dialogue state and available action templates.

        :param state: The current dialogue state.
        :type state: dict

        :param templates_slots: The available different combinations of slots that can be filled by templates.
        :type templates_slots: list[tuple]

        :return: The index of the selected action template, or None if no action is available.
        :rtype: int or None
        """
        
        filled = []
        for slot, value in state.items():
          if value is not None:
            filled.append(slot)

        filled = tuple(sorted(filled))

        top_idx = -1
        for idx, slots in enumerate(templates_slots):
          if all(element not in slots for element in filled):
            top_idx = idx
            break

        action = None
        if top_idx != -1:
          action = top_idx

        return action

    def __str__(self):
      """
      Return a string representation of the PriorityDP policy.

      :return: A string representation of the PriorityDP policy.
      :rtype: str
      """
      string = ''
      string = string + '\nPriorityDP'
      return string


class RandomDP:
    """
    Dialogue Policy Optimizer that selects the next action randomly.

    This policy selects the action randomly from the available action templates
    that do not conflict with the slots already filled in the dialogue state.

    """
    def get_action(self, state, templates_slots):
        """
        Get the next action based on the given dialogue state and available action templates.

        :param state: The current dialogue state.
        :type state: dict

        :param templates_slots: The available different combinations of slots that can be filled by templates.
        :type templates_slots: list[tuple]

        :return: The index of the selected action template, or None if no action is available.
        :rtype: int or None
        """
        filled = []
        for slot, value in state.items():
          if value is not None:
            filled.append(slot)

        filled = tuple(sorted(filled))

        candidates = []
        for idx, slots in enumerate(templates_slots):
          if all(element not in slots for element in filled):
            candidates.append(idx)

        # print('candidates', candidates)

        max_len = len(candidates)

        action = None

        if max_len > 0:
          action = candidates[random.randint(0, max_len-1)]

        return action

    def __str__(self):
      """
      Return a string representation of the RandomDP policy.

      :return: A string representation of the RandomDP policy.
      :rtype: str
      """
      string = ''
      string = string + '\nRandomDP'
      return string


class TemplateBasedNLG:
    """
    Natural Language Generation module that generates responses based on predefined templates.

    This module uses a set of predefined templates containing utterances and corresponding system acts.
    Given an index, it generates the corresponding system utterance and system act.

    :param templates: The predefined templates for generating responses.
    :type templates: list[dict]
    """

    def __init__(self, templates):
      self.templates = templates
      self.templates_slots = []

      for template in templates:
        self.templates_slots.append(tuple(sorted(template['slots'])))

    def get_templates(self):
      """
      Get the predefined templates.

      :return: The predefined templates.
      :rtype: list[dict]
      """
      return self.templates

    def get_templates_slots(self):
      """
      Get the slots associated with the predefined templates.

      :return: The slots associated with the predefined templates.
      :rtype: list[tuple]
      """
      return self.templates_slots

    def generate(self, idx):
      """
      Generate a response based on the given index.

      :param idx: The index of the template to generate a response from.
      :type idx: int

      :return: The generated system utterance and system act.
      :rtype: tuple[str, list] or None
      """

      if idx < 0 or idx >= len(self.templates):
        return None, None

      return self.templates[idx]['utterance'], self.templates[idx]['system_act']

    def __str__(self):
      """
      Return a string representation of the TemplateBasedNLG module.

      :return: A string representation of the TemplateBasedNLG module.
      :rtype: str
      """
      string = ''
      string = string + '\ntemplates: ' + str(self.templates)
      string = string + '\ntemplates_slots: ' + str(self.templates_slots)
      return string


"""
This Module has the training functions for TRIPPY.
"""

import torch
import torch.nn as nn
from tqdm import tqdm

from botiverse.models.TRIPPY.utils import AverageMeter


def span_loss_fn(start_logits, end_logits, targets_start, targets_end, ignore_idx):
  """
  Compute the span loss.

  :param start_logits: The start logits.
  :type start_logits: torch.Tensor
  :param end_logits: The end logits.
  :type end_logits: torch.Tensor
  :param targets_start: The start targets.
  :type targets_start: torch.Tensor
  :param targets_end: The end targets.
  :type targets_end: torch.Tensor
  :param ignore_idx: The index to ignore in the loss calculation.
  :type ignore_idx: int
  :return: The span loss.
  :rtype: torch.Tensor
  """
  l1 = nn.CrossEntropyLoss(ignore_index=ignore_idx, reduction='none')(start_logits, targets_start)
  l2 = nn.CrossEntropyLoss(ignore_index=ignore_idx, reduction='none')(end_logits, targets_end)
  return (l1 + l2) / 2.0

def oper_loss_fn(oper_logits, oper_labels, ignore_idx):
  """
  Compute the operation loss.

  :param oper_logits: The operation logits.
  :type oper_logits: torch.Tensor
  :param oper_labels: The operation labels.
  :type oper_labels: torch.Tensor
  :param ignore_idx: The index to ignore in the loss calculation.
  :type ignore_idx: int
  :return: The operation loss.
  :rtype: torch.Tensor
  """
  l = nn.CrossEntropyLoss(ignore_index=ignore_idx, reduction='none')(oper_logits, oper_labels)
  return l

def refer_loss_fn(refer_logits, refer_labels, ignore_idx):
  """
  Compute the refer loss.

  :param refer_logits: The refer logits.
  :type refer_logits: torch.Tensor
  :param refer_labels: The refer labels.
  :type refer_labels: torch.Tensor
  :param ignore_idx: The index to ignore in the loss calculation.
  :type ignore_idx: int
  :return: The refer loss.
  :rtype: torch.Tensor
  """
  l = nn.CrossEntropyLoss(ignore_index=ignore_idx, reduction='none')(refer_logits, refer_labels)
  return l

def train(data_loader, model, optimizer, device, scheduler, n_slots, ignore_idx, oper2id):
  """
  Perform the training loop for a model on the given data.

  :param data_loader: The data loader providing the training batches.
  :type data_loader: DataLoader

  :param model: The model to be trained.
  :type model: nn.Module

  :param optimizer: The optimizer used to update the model's parameters.
  :type optimizer: Optimizer

  :param device: The device (e.g., CPU or GPU) on which the training will be performed.
  :type device: torch.device

  :param scheduler: The scheduler for adjusting the learning rate during training.
  :type scheduler: _LRScheduler

  :param n_slots: The number of slots in the task.
  :type n_slots: int

  :param ignore_idx: The index to ignore during loss computation.
  :type ignore_idx: int

  :param oper2id: A dictionary mapping operation names to their corresponding IDs.
  :type oper2id: dict
  """

  model.train()

  losses = AverageMeter()

  tk0 = tqdm(data_loader)
  for i, batch in enumerate(tk0):

    ids = batch['ids'].to(device)
    mask = batch['mask'].to(device)
    token_type_ids = batch['token_type_ids'].to(device)
    spans_start = batch['spans_start'].to(device)
    spans_end = batch['spans_end'].to(device)
    opers = batch['opers'].to(device)
    refer = batch['refer'].to(device)
    inform_aux_features = batch['inform_aux_features'].to(device)
    ds_aux_features = batch['ds_aux_features'].to(device)

    optimizer.zero_grad()
    slots_start_logits, slots_end_logits, slots_oper_logits, slots_refer_logits = model(ids=ids,
                                                                                        mask=mask,
                                                                                        token_type_ids=token_type_ids,
                                                                                        inform_aux_features=inform_aux_features,
                                                                                        ds_aux_features=ds_aux_features)

    batch_loss = 0.0

    for slot in range(n_slots):

      oper_loss = oper_loss_fn(slots_oper_logits[slot], opers[:,slot], ignore_idx)

      span_loss = span_loss_fn(slots_start_logits[slot],
                               slots_end_logits[slot],
                               spans_start[:,slot],
                               spans_end[:,slot],
                               ignore_idx)
      token_is_pointable = (spans_start[:,slot] > 0).float()
      span_loss *= token_is_pointable

      refer_loss = refer_loss_fn(slots_refer_logits[slot], refer[:,slot], ignore_idx)
      token_is_referrable = (opers[:,slot] == oper2id['refer']).float()
      refer_loss *= token_is_referrable

      total_loss = 0.8 * oper_loss + 0.1 * span_loss + 0.1 * refer_loss

      batch_loss += total_loss.sum()

    # batch_loss /= n_slots
    batch_loss.backward()
    torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)
    optimizer.step()
    scheduler.step()
    losses.update(batch_loss.item(), ids.size(0))

    tk0.set_postfix(loss=losses.avg)


import numpy as np

def split_data(X,y, val_ratio):
    '''
    Split the data into training and validation sets.
    '''
    N = X.shape[0]
    # randomly shuffle the data with numpy's permutation function.
    index = np.random.permutation(N)
    X, y = X[index], y[index]
    # split
    V = int(N * val_ratio)
    X_t, y_t = X[:N - V], y[:N - V]
    X_v, y_v = X[N - V:], y[N - V:]
    return X_t, y_t, X_v, y_v


def batchify(x_data, y_data, batch_size):
    '''
    Given x_data of shape (N, d) and y_data of shape (N) return x_data of shape (B, N//B, d, 1) and y_data of shape (B, N//B, K, 1).
    '''
    N = x_data.shape[0]
    # shuffle the data with numpy's permutation function.
    index = np.random.permutation(N)
    x_data, y_data = x_data[index], y_data[index]
    
    # add a trailing dimension to x_data
    x_data = x_data[..., np.newaxis]
    
    # make y_data a one-hot vector
    u = len(np.unique(y_data))
    y_data = np.array([np.identity(u)[:,[y]] for y in y_data])           #y is a one-hot column vector.
    
    # truncate the data to be divisible by the batch size
    x_data = x_data[:N - N%batch_size]
    y_data = y_data[:N - N%batch_size]
    
    # batchify the data using numpy
    x_data = np.split(x_data, N//batch_size)
    y_data = np.split(y_data, N//batch_size)
    
    return np.array(x_data), np.array(y_data)


"""
This Module has the evaluation functions for TRIPPY.
"""
import torch
import numpy as np
from sklearn.metrics import f1_score
import copy
from tqdm import tqdm

from botiverse.models.TRIPPY.utils import normalize, is_included, included_with_label_maps, create_span_output


def get_informed_value(value, target, label_maps, multiwoz):
  """
  Get the informed value based on the value and target, taking into account label maps.

  :param value: The original value.
  :type value: str
  :param target: The target value to compare with.
  :type target: str
  :param label_maps: The mapping of slot values to their variants.
  :type label_maps: dict
  :return: The informed value.
  :rtype: str
  """

  informed = False
  informed_value = value

  value = ' '.join(normalize(value, multiwoz))
  target = ' '.join(normalize(target, multiwoz))

  if value == target or is_included(value, target) or is_included(target, value):
    informed = True
  if value in label_maps:
    informed = included_with_label_maps(target, value, label_maps)
  elif target in label_maps:
    informed = included_with_label_maps(value, target, label_maps)
  if informed: informed_value = target

  return informed_value


def eval(raw_data, data, model, device, n_slots, slot_list, label_maps, oper2id, multiwoz):
  """
  Evaluate the model on the given data.

  :param raw_data: The raw data.
  :type raw_data: list
  :param data: The processed data.
  :type data: list
  :param model: The model to evaluate.
  :type model: nn.Module
  :param device: The device to run the evaluation on.
  :type device: torch.device
  :param n_slots: The number of slots.
  :type n_slots: int
  :param slot_list: The list of slots.
  :type slot_list: list
  :param label_maps: The mapping of slot values to their variants.
  :type label_maps: dict
  :param oper2id: The mapping of operations to their IDs.
  :type oper2id: dict
  :return: The evaluation metrics.
  :rtype: tuple
  """

  model.eval()

  # normalize label_maps
  label_maps_tmp = {}
  for v in label_maps:
      label_maps_tmp[' '.join(normalize(v, multiwoz))] = [' '.join(normalize(nv, multiwoz)) for nv in label_maps[v]]
  label_maps = label_maps_tmp


  # metrics
  Y_true, Y_pred = [], []
  per_slot_acc = {slot:[] for slot in slot_list}
  joint_goal_acc = []

  # state
  pred_last_state = {}
  pre_dialogue_idx = -1

  # debugging
  states = []
  sentences = []
  indices = []
  prev_idx = -1

  with torch.no_grad():

    for raw_turn, turn in tqdm(zip(raw_data, data), total=len(raw_data)):

      ids = torch.tensor(turn.ids, dtype=torch.long).unsqueeze(0).to(device)
      mask = torch.tensor(turn.mask, dtype=torch.long).unsqueeze(0).to(device)
      token_type_ids = torch.tensor(turn.token_type_ids, dtype=torch.long).unsqueeze(0).to(device)
      inform_aux_features = torch.tensor(turn.inform_aux_features, dtype=torch.float).unsqueeze(0).to(device)
      # ds_aux_features = torch.tensor(turn.ds_aux_features, dtype=torch.float).unsqueeze(0).to(device)
      input_tokens = ' '.join(turn.input_tokens)
      padding_len = turn.padding_len
      turn_idx = raw_turn.turn_idx
      dialogue_idx = raw_turn.dial_idx
      current_state = turn.cur_state
      inform_mem = raw_turn.inform_mem
      opers = turn.opers

      # new dialogue reset the state and the state auxiliary features
      if turn_idx == 0 or dialogue_idx != pre_dialogue_idx:
        pred_last_state = {}
        ds_aux_features = torch.zeros((1, n_slots)).to(device)

      # get model outputs
      slots_start_logits, slots_end_logits, slots_oper_logits, slots_refer_logits = model(ids=ids,
                                                                                          mask=mask,
                                                                                          token_type_ids=token_type_ids,
                                                                                          inform_aux_features=inform_aux_features,
                                                                                          ds_aux_features=ds_aux_features)

      # update the predicted state of each slot
      pred_current_state = pred_last_state.copy()
      for slot_idx, slot in enumerate(slot_list):

        # get the predicted operation
        pred_oper = slots_oper_logits[slot_idx][0].argmax(dim=-1).item()

        # keep track of operations for f1 score
        Y_pred.append(pred_oper)
        Y_true.append(oper2id[opers[slot_idx]])

        # update the slot based on the operation
        if pred_oper == oper2id['carryover']: # carryover
          continue
        elif pred_oper == oper2id['dontcare']: # dontcare
          pred_current_state[slot] = 'dontcare'
        elif pred_oper == oper2id['update']: # update
          pred_current_state[slot] = create_span_output(slots_start_logits[slot_idx][0].cpu().detach().numpy(),
                                                        slots_end_logits[slot_idx][0].cpu().detach().numpy(),
                                                        padding_len,
                                                        input_tokens)
        elif pred_oper == oper2id['refer']: # refer
          refered_slot = slots_refer_logits[slot_idx][0].argmax(dim=-1).item()
          if refered_slot != n_slots and slot_list[refered_slot] in pred_last_state:
            pred_current_state[slot] = pred_last_state[slot_list[refered_slot]]
        elif pred_oper == oper2id['yes']: # yes
          pred_current_state[slot] = 'yes'
        elif pred_oper == oper2id['no']: # no
          pred_current_state[slot] = 'no'
        elif pred_oper == oper2id['inform']: # inform
          if slot in inform_mem:
            pred_current_state[slot] = '§§' + inform_mem[slot][0]

      # update the state auxiliary features
      for slot_idx, slot in enumerate(slot_list):
          ds_aux_features[0, slot_idx] = 1 if slot in pred_current_state else 0

      # calculate accuracy
      joint = True
      for slot_idx, slot in enumerate(slot_list):

        # if not present in both
        if slot not in current_state and slot not in pred_current_state:
          per_slot_acc[slot].append(1.0)
          continue

        # if slot only in one of them then mark as 0
        if (slot in current_state and slot not in pred_current_state) or (slot not in current_state and slot in pred_current_state):
          joint = False
          per_slot_acc[slot].append(0.0)
          continue

        # get values
        val = current_state[slot]
        pred_val = pred_current_state[slot]

        # normalize values
        val = ' '.join(normalize(val, multiwoz))
        pred_val = ' '.join(normalize(pred_val, multiwoz))

        # handle inform
        if pred_val[0:3] == "§§ ":
          if pred_val[3:] != 'none':
              pred_val = get_informed_value(pred_val[3:], val, label_maps, multiwoz)
        elif pred_val[0:2] == "§§":
            if pred_val[2:] != 'none':
                pred_val = get_informed_value(pred_val[2:], val, label_maps, multiwoz)

        # match
        if pred_val == val:
          per_slot_acc[slot].append(1.0)
        elif val != 'none' and val != 'dontcare' and val != 'true' and val != 'false' and val in label_maps:
          no_match = True
          for variant in label_maps[val]:
              if variant == pred_val:
                  no_match = False
                  break
          if no_match:
              per_slot_acc[slot].append(0.0)
              joint = False
          else:
              per_slot_acc[slot].append(1.0)
        else:
            per_slot_acc[slot].append(0.0)
            joint = False

      # append joint
      joint_goal_acc.append(1.0 if joint else 0.0)


      # update vars for next turn
      pred_last_state = pred_current_state.copy()
      pre_dialogue_idx = dialogue_idx

#       # debugging
#       if per_slot_acc['attraction-name'][-1] < 0.99 and prev_idx != dialogue_idx:
#         print('dialogue_idx', dialogue_idx)
#         print('turn_idx', turn_idx)
#         print('pred_state', dict(sorted(pred_current_state.items())))
#         print('cur_state', dict(sorted(current_state.items())))
#         print('input tok', input_tokens)
#         print('inform_mem', inform_mem)
#         print('inform aux', inform_aux_features)
#         print('oper', opers[slot_list.index('attraction-name')])
#         prev_idx = dialogue_idx

      # debugging
      if joint == False:
        states.append((copy.deepcopy(pred_current_state), current_state))
        sentences.append(input_tokens)
        indices.append((dialogue_idx, turn_idx))

  # debugging
  # prev = ""
  # for i in range(len(states)):
  #   if prev == indices[i][0] or len(states[i][0]) != len(states[i][1]):
  #     continue
  #   if 'attraction-name' not in states[i][1]:# and 'restaurant-name' not in states[i][1]:
  #     continue
  #   prev = indices[i][0]
  #   print(dict(sorted(states[i][0].items())))
  #   print(dict(sorted(states[i][1].items())))
  #   print(sentences[i])
  #   print(indices[i])
  #   print("\n")

  # calculate per slot accuracy
  per_slot_acc = {slot: np.mean(acc) for slot, acc in per_slot_acc.items()}

  # calculate f1 scores
  macro_f1_score = f1_score(Y_true, Y_pred, average='macro')
  all_f1_score = f1_score(Y_true, Y_pred, average=None)

  return np.mean(joint_goal_acc), per_slot_acc, macro_f1_score, all_f1_score


#!/usr/bin/env python
# coding: utf-8

# # <font color="cyan">Support Vector Machine </font> From Scratch Implementation
# 
# In this notebook, we shall demonstrate implementing multiclass SVM from scratch using the dual formulation and quadratic programming.

# In[2]:


'''
This module implements and provides an interface for the multiclass Support Vector Machine (SVM) algorithm implemented from scratch.
'''
import numpy as np
from scipy.spatial import distance
import cvxopt
import matplotlib.pyplot as plt
from sklearn.datasets import make_classification
import copy
import pickle


# ### SVM Class

# In[3]:


class SVM:
    '''
    This class implements the kernelized soft-margin Support Vector Machine algorithm.
    '''
    linear = lambda x, xࠤ , c=0: x @ xࠤ .T
    polynomial = lambda x, xࠤ , Q=5: (1 + x @ xࠤ.T)**Q
    rbf = lambda x, xࠤ , γ=10: np.exp(-γ * distance.cdist(x, xࠤ, 'sqeuclidean'))
    kernel_funs = {'linear': linear, 'polynomial': polynomial, 'rbf': rbf}
    
    def __init__(self, kernel='rbf', C=1, k=2):
        '''
        Initialize an instance of the SVM model.
        :param kernel: The kernel function to use. Can be 'linear', 'polynomial', or 'rbf'.
        :param C: The regularization parameter.
        :param k: The hyperparameter for the polynomial and rbf kernels. Ignored for the linear kernel.
        '''
        self.kernel_str = kernel
        self.kernel = SVM.kernel_funs[kernel]
        self.C = C
        self.k = k
        self.X, y = None, None
        self.αs = None
        # for multi-class classification
        self.multiclass = False
        self.clfs = []                                  

SVMClass = lambda func: setattr(SVM, func.__name__, func) or func


# #### <font color="white">Solve the Dual Optimization Problem </font>

# We want to find the vector of Lagrange multipliers $\alpha = (\alpha_1, \alpha_2, ..., \alpha_n)$ that maximizes the following dual objective function:
# 
# $$
# \max_\alpha [ \sum_{i=1}^{n} \alpha_i - \frac{1}{2} \sum_{i=1}^{n} \sum_{j=1}^{n} \alpha_i \alpha_j y_i y_j (z_iz_j^T) ]
# \quad
# \text{subject to} \quad 
# 
# \sum_{i=1}^{n} \alpha_i y_i = 0, \;  0 \leq \alpha_i \leq C, \; \forall i = 1, 2, ..., n
# $$
# 
# where $z_i = \phi(x_i)$ and $z_j = \phi(x_j)$ are higher dimensional representations of $x_i$ and $x_j$ respectively given by the transform $\phi$.
# 
# <br>
# In matrix form, this becomes
# 
# $$\max_\alpha [ 1^Tα - \frac{1}{2} * α^T (YY^T * K) α ] \quad
# \text{subject to} \quad
# y^T \alpha = 0, \; 0 \leq \alpha_i \leq C, \; \forall i = 1, 2, \ldots, n
# $$
# where $K[i, j] = z_i^T z_j$ and $X$, $y$ are $(N,d)$ and $(N,1)$ dimensional training data matrices respectively.
# <br>
# CVXOPT solves
# 
# $$
# \min_x \frac{1}{2} x^T P x + q^T x \quad
# \text{subject to} \;  A x = b, \;G x ≤ h 
# $$
# 
# Hence, we have
# <br>
# $ x = \alpha $
# <br>
# $ P = YY^T * K $
# <br>
# $ q = -1^T $
# <br><br>
# and for the constraints
# <br>
# $ A = y^T $
# <br>
# $ b = 0 $
# <br>
# $ G = \begin{bmatrix} -I \\ I \end{bmatrix} $
# <br>
# $ h = \begin{bmatrix} 0 \\ C \end{bmatrix} $

# In[4]:


@SVMClass
def fit(self, X, y, eval_train=False):
    '''
    Fit the SVM model to the given data with N training examples and d features.
    :param X: The training data arranged as a float numpy array of shape (N, d)
    :param y: The training labels arranged as a numpy array of shape (N,) where each element is in {-1, 1} or {0, 1}.
    :return: None
    
    :note: This function assume C is not too small; otherwise, it becomes hard to distinguish support vectors.
    '''
    if len(np.unique(y)) > 2:
        self.multiclass = True
        return self.multi_fit(X, y, eval_train)
    
    if set(np.unique(y)) == {0, 1}: y[y == 0] = -1
    self.y = y.reshape(-1, 1).astype(np.double) # Has to be a column vector
    self.X = X
    m = X.shape[0]
    
    # compute the kernel over all possible pairs of (x, x') in the data
    self.K = self.kernel(X, X, self.k)
    
    # For 1/2 x^T P x + q^T x
    P = cvxopt.matrix(self.y @ self.y.T * self.K)
    q = cvxopt.matrix(-np.ones((m, 1)))
    
    # For Ax = b
    A = cvxopt.matrix(self.y.T)
    b = cvxopt.matrix(np.zeros(1))

    # For Gx <= h
    G = cvxopt.matrix(np.vstack((-np.identity(m),
                                 np.identity(m))))
    h = cvxopt.matrix(np.vstack((np.zeros((m,1)),
                                 np.ones((m,1)) * self.C)))

    # Solve    
    cvxopt.solvers.options['show_progress'] = False
    sol = cvxopt.solvers.qp(P, q, G, h, A, b)
    self.αs = np.array(sol["x"])
        
    # Maps into support vectors
    self.is_sv = ((self.αs > 1e-3) & (self.αs <= self.C)).squeeze()
    self.margin_sv = np.argmax((1e-3 < self.αs) & (self.αs < self.C - 1e-2))
    
    if eval_train:  print(f"Model finished training with accuracy {self.evaluate(X, y)}")


# #### Multiclass Case with OVR

# In this, we train $K$ binary classifiers, where $K$ is the number of classes where each classifier perceives one class as +1 and all other classes as -1.

# In[5]:


@SVMClass
def multi_fit(self, X, y, eval_train=False):
    '''
    Fit k classifier for k classes.
    :param X: The training data arranged as a float numpy array of shape (N, d)
    :param y: The training labels arranged as a numpy array of shape (N,) where each element is an integer between 0 and k-1
    :return: None
    '''
    self.k = len(np.unique(y))      # number of classes
    # for each pair of classes
    for i in range(self.k):
        # get the data for the pair
        Xs, Ys = X, copy.copy(y)
        # change the labels to -1 and 1
        Ys[Ys!=i], Ys[Ys==i] = -1, +1
        # fit the classifier
        clf = SVM(kernel=self.kernel_str, C=self.C, k=self.k)
        clf.fit(Xs, Ys)
        # save the classifier
        self.clfs.append(clf)
    if eval_train:  print(f"Model finished training with accuracy {self.evaluate(X, y)}")


# #### SVM Prediction

# Given a new data point $x$, we predict its label $y$ by computing:
# 
# $$g(x) = \sum_{i=1}^{n} \alpha_i y_i k(x_i, x) + b$$
# 
# where $k(x_i, x)$ is the kernel function and $b= y_s - \sum_{i=1}^{n} \alpha_i y_i k(x_i, x_s)$ is the bias term. $s$ is the index of any margin support vector.
# 
# 

# In[6]:


@SVMClass
def predict(self, X_t):
    '''
    Predict the labels for given test data.
    :param X_t: The test data arranged as a float numpy array of shape (N, d)
    :param return_score: If True, return the score instead of the label.
    :return: The predicted labels arranged as a numpy array of shape (N,) where each element is in {-1, 1} or {0, 1} unless return_score is True in which case the SVM score is returned.
    '''
    if self.multiclass: return self.multi_predict(X_t)
    xₛ, yₛ = self.X[self.margin_sv, np.newaxis], self.y[self.margin_sv]
    αs, y, X= self.αs[self.is_sv], self.y[self.is_sv], self.X[self.is_sv]

    b = yₛ - np.sum(αs * y * self.kernel(X, xₛ, self.k), axis=0)
    score = np.sum(αs * y * self.kernel(X, X_t, self.k), axis=0) + b
    return np.sign(score).astype(int), score


@SVMClass
def evaluate(self, X,y):
    '''
    Compare the predicted labels for given test data y with the actual labels by passing X to model.
    :param X: The test data arranged as a float numpy array of shape (N, d)
    :param y: The test labels arranged as a numpy array of shape (N,) where each element is an integer between 0 and k-1
    '''
    outputs, _ = self.predict(X)
    accuracy = np.sum(outputs == y) / len(y)
    
    return round(accuracy, 2)


# #### Multiclass Case with OVR

# Each classifier compute the score of a class against all other classes. The class with the highest score is the predicted class.

# In[7]:


@SVMClass
def multi_predict(self, X):
    '''
    Predict the labels for given test data.
    :param X: The test data arranged as a float numpy array of shape (N, d)
    :return: The predicted labels arranged as a numpy array of shape (N,) where each element is an integer between 0 and k-1.
    '''
    # get the predictions from all classifiers
    preds = np.zeros((X.shape[0], self.k))
    for i, clf in enumerate(self.clfs):
        _, preds[:, i] = clf.predict(X)
    

    # transform the predictions into probabilities using softmax
    preds = np.exp(preds) / np.sum(np.exp(preds), axis=1, keepdims=True)    
    
    # get the argmax and the corresponding score
    return np.argmax(preds, axis=1), np.max(preds, axis=1)


# #### Saving and Loading

# In[8]:


@SVMClass
def save(self, path):
    '''
    Save the model to the given path.
    :param path: The path to save the model to.
    '''
    with open(path, 'wb') as f:
        pickle.dump(self, f)
        

@staticmethod
@SVMClass
def load(path):
    '''
    load the model from the given path.
    :param path: The path to load the model from.
    '''
    with open(path, 'rb') as f:
        return pickle.load(f)
        


# ### Example

# In[9]:


# if running from notebook
if __name__ == '__main__':
    get_ipython().system('jupyter nbconvert --to script SVM.ipynb')

